# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from Bio import SeqIO
import requests
from io import StringIO
from hashlib import md5


def get_protein_sequence(uniprot_id: str) -> str:
    """
    This function takes a UniProt ID as input and returns the corresponding sequence record.

    Parameters:
        uniprot_id (str): The UniProt identifier of the protein of interest

    Returns:
        list: List with results as Bio.SeqRecord.SeqRecord object

    Example:
        ASMT_representations = uniprot.get_protein_sequence('P46597')
        sequence_string = str(ASMT_representations[0].seq)
    """
    base_url = "http://www.uniprot.org/uniprot/"
    current_url = base_url + uniprot_id + ".fasta"
    response = requests.post(current_url)
    c_data = "".join(response.text)

    seq = StringIO(c_data)
    p_seq = list(SeqIO.parse(seq, "fasta"))
    return p_seq


def get_uniprot_id(sequence: str) -> str:
    """
    This function takes in a protein sequence string and returns the corresponding UniProt ID.

    Parameters:
        sequence (str): The protein sequence string

    Returns:
        str: The UniProt ID of the protein, if it exists in UniProt. None otherwise.
    """
    h = md5(sequence.encode()).digest().hex()
    requestURL = f"https://www.ebi.ac.uk/proteins/api/proteins?offset=0&size=100&md5={h}"  # 5e2c446cc1c54ee4406b9f6683b7f98d
    r = requests.get(requestURL, headers={"Accept": "application/json"})

    if not r.ok:
        return None
    data = r.json()

    if len(data) == 0:
        return None
    else:
        return data[0]["accession"]



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from sklearn.model_selection import GridSearchCV
from sklearn.ensemble import RandomForestRegressor
from scipy.stats import pearsonr
from sklearn.neighbors import KNeighborsRegressor
import pandas as pd
from sklearn.svm import SVR
import numpy


def knnr_grid_search(
    Xs_train: numpy.ndarray,
    Xs_test: numpy.ndarray,
    ys_train: list,
    ys_test: list,
    param_grid: dict = None,
    verbose: int = 1,
):
    """
    Performs a KNN regressor grid search using 5 fold cross validation.

    Parameters:
    -----------
        Xs_train, Xs_test (numpy.ndarray): train and test values for training
        ys_train, ys_test (list): y values for train and test
        param_grid (dict): parameter grid for model
        verbose (int): Type of information printing during run

    Returns:
    --------
        Returns the best performing model of grid search, the test R squared value, correlation coefficient,
        p-value of the fit and a dataframe containing the fit information.
    """
    # Instantiate the model
    knnr = KNeighborsRegressor()

    if param_grid is None:
        param_grid = {
            "n_neighbors": [3, 5, 7],
            "weights": ["uniform", "distance"],
            "algorithm": ["ball_tree", "kd_tree", "brute"],
            "leaf_size": [10, 15, 20],
            "p": [1, 2, 3],
        }
    # Create a GridSearchCV object and fit the model
    # grid_search = GridSearchCV(estimator=knnr, param_grid=param_grid, cv=5, n_jobs=-1, verbose=1)
    grid_search = GridSearchCV(
        estimator=knnr,
        param_grid=param_grid,
        scoring="r2",
        cv=5,
        verbose=verbose,
        n_jobs=-1,  # use all available cores
    )

    grid_search.fit(Xs_train, ys_train)
    test_r2 = grid_search.score(Xs_test, ys_test)
    predictions = grid_search.best_estimator_.predict(Xs_test)
    corr_coef, p_value = pearsonr(predictions, ys_test)

    # Print the best hyperparameters and the best score
    if verbose is not None:
        print("Best hyperparameters: ", grid_search.best_params_)
        print("Best score: ", grid_search.best_score_)
        print("Test R^2 score: ", test_r2)
        print("Correlation coefficient: {:.2f}".format(corr_coef))
        print("p-value: {:.4f}".format(p_value))

    return (
        grid_search.best_estimator_,
        test_r2,
        corr_coef,
        p_value,
        pd.DataFrame.from_dict(grid_search.cv_results_),
    )


def rfr_grid_search(
    Xs_train: numpy.ndarray,
    Xs_test: numpy.ndarray,
    ys_train: list,
    ys_test: list,
    param_grid: dict = None,
    verbose: int = 1,
):
    """
    Performs a Random Forrest regressor grid search using 5 fold cross validation.

    Parameters:
    -----------
        Xs_train, Xs_test (numpy.ndarray): train and test values for training
        ys_train, ys_test (list): y values for train and test
        param_grid (dict): parameter grid for model
        verbose (int): Type of information printing during run

    Returns:
    --------
        Returns the best performing model of grid search, the test R squared value, correlation coefficient,
        p-value of the fit and a dataframe containing the fit information.
    """
    # Instantiate the model
    rfr = RandomForestRegressor(random_state=42)

    if param_grid is None:
        param_grid = {
            "n_estimators": [20, 50, 100, 200],
            "criterion": ["squared_error", "absolute_error"],
            "max_features": ["sqrt", "log2"],
            "max_depth": [5, 10, 15],
            "min_samples_split": [2, 5, 10],
            "min_samples_leaf": [1, 4],
        }

    # Create a GridSearchCV object and fit the model
    grid_search = GridSearchCV(
        estimator=rfr,
        param_grid=param_grid,
        scoring="r2",
        cv=5,
        verbose=verbose,
        n_jobs=-1,  # use all available cores
    )
    grid_search.fit(Xs_train, ys_train)

    # Evaluate the performance of the model on the test set
    test_r2 = grid_search.score(Xs_test, ys_test)
    predictions = grid_search.best_estimator_.predict(Xs_test)
    corr_coef, p_value = pearsonr(predictions, ys_test)

    if verbose is not None:
        # Print the best hyperparameters and the best score
        print("Best hyperparameters: ", grid_search.best_params_)
        print("Best score: ", grid_search.best_score_)
        print("Test R^2 score: ", test_r2)
        print("Correlation coefficient: {:.2f}".format(corr_coef))
        print("p-value: {:.4f}".format(p_value))

    return (
        grid_search.best_estimator_,
        test_r2,
        corr_coef,
        p_value,
        pd.DataFrame.from_dict(grid_search.cv_results_),
    )


def svr_grid_search(
    Xs_train: numpy.ndarray,
    Xs_test: numpy.ndarray,
    ys_train: list,
    ys_test: list,
    param_grid: dict = None,
    verbose: int = 1,
):
    """
    Performs a Support Vector regressor grid search using 5 fold cross validation.

    Parameters:
    -----------
        Xs_train, Xs_test (numpy.ndarray): train and test values for training
        ys_train, ys_test (list): y values for train and test
        param_grid (dict): parameter grid for model
        verbose (int): Type of information printing during run

    Returns:
    --------
        Returns the best performing model of grid search, the test R squared value, correlation coefficient,
        p-value of the fit and a dataframe containing the fit information.
    """
    # Instantiate the model
    svr = SVR()

    if param_grid is None:
        param_grid = {
            "C": [0.1, 1, 2, 5, 10, 100, 200, 400],
            "gamma": ["scale"],
            "kernel": ["linear", "poly", "rbf", "sigmoid"],
            "degree": [3],
        }

    # Create a GridSearchCV object and fit the model
    grid_search = GridSearchCV(
        estimator=svr,
        param_grid=param_grid,
        scoring="r2",
        cv=5,
        verbose=verbose,
        n_jobs=-1,  # use all available cores
    )
    grid_search.fit(Xs_train, ys_train)

    # Evaluate the performance of the model on the test set
    test_r2 = grid_search.score(Xs_test, ys_test)
    predictions = grid_search.best_estimator_.predict(Xs_test)
    corr_coef, p_value = pearsonr(predictions, ys_test)

    if verbose is not None:
        # Print the best hyperparameters and the best score
        print("Best hyperparameters: ", grid_search.best_params_)
        print("Best score: ", grid_search.best_score_)
        print("Test R^2 score: ", test_r2)
        print("Correlation coefficient: {:.2f}".format(corr_coef))
        print("p-value: {:.4f}".format(p_value))

    return (
        grid_search.best_estimator_,
        test_r2,
        corr_coef,
        p_value,
        pd.DataFrame.from_dict(grid_search.cv_results_),
        grid_search.best_params_,
    )



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

import os
import sys
import warnings
from typing import Union

import biotite.structure
import pandas as pd
import torch
import biotite
import biotite.structure.io as strucio

import proteusAI.ml_tools.esm_tools.esm_tools as esm_tools
import proteusAI.struc as struc

current_path = os.path.dirname(os.path.abspath(__file__))
root_path = os.path.join(current_path, "..")
sys.path.append(root_path)
folder_path = os.path.dirname(os.path.realpath(__file__))
USR_PATH = os.path.join(folder_path, "../../../usrs")

model_dict = {
    "rf": "Random Forrest",
    "knn": "KNN",
    "svm": "SVM",
    "vae": "VAE",
    "esm2": "ESM-2",
    "esm1v": "ESM-1v",
}


class Protein:
    """
    The protein object contains information about a single protein,
    such as its sequence, name, and its mathematical representations.

    Attributes:
        name (str): Name/id of the protein.
        seq (str): Protein sequence.
        rep (list): List of available representations.
        rep_path (str): Path to representations directory.
        design (str): Output of design.
    """

    def __init__(
        self,
        name: Union[str, None] = None,
        seq: Union[str, None] = None,
        struc: Union[str, biotite.structure.AtomArray, None] = None,
        reps: Union[list, tuple] = [],
        user: Union[str, None] = "guest",
        y=None,
        y_pred=None,
        y_sigma=None,
        acq_score=None,
        source: Union[str, None] = None,
        fname: Union[str, None] = None,
    ):
        """
        Initialize a new protein object.

        Args:
            name (str): Name/id of the protein.
            seq (str): Protein sequence.
            struc (AtomArray): Protein structure.
            reps (list): List of available representations.
            user (str): Path to the user. Will create one if the path does not exist. Default guest.
            y (float, int, str): Label for the protein.
            y_pred (float, int, str): Predicted y_value.
            y_sigma (float, int, str): Predicted y_value.
            acq_score (float): acquisition score.
            source (str, or data): Source of data, either a file or a data package created from a diversification step.
            fname (str): Only relevant for the app - provides the real file name instead of temporary file name from shiny.

        Parameters:

        """

        # Arguments
        self.name = name
        self.seq = seq
        self.reps = list(reps)
        self.struc = struc
        self.source = source
        self.y = y
        self.y_pred = y_pred
        self.y_sigma = y_sigma
        self.acq_score = acq_score
        self.fname = fname
        self.user = os.path.join(USR_PATH, user)

        # Parameters
        self.pdb_file = None
        self.fasta_file = None
        self.reps = []
        self.source_path = None
        self.rep_path = None
        self.struc_path = None
        self.design = None
        self.chains = []
        self.zs_path = None
        self.class_dict = None

        # Create user if user does not exist
        if not os.path.exists(self.user):
            self.initialize_user()

        # Initialize library from file or from inheritance
        if isinstance(self.source, str):
            self.init_from_file()
        elif self.source is not None:
            self.init_from_inheritance()

    def __str__(self):
        if self.struc is not None:
            struc_loaded = "loaded"  # noqa: F841
        return f"proteusAI.Protein():\n____________________\nname\t: {self.name}\nseq\t: {self.seq}\nrep\t: {self.reps}\ny:\t{self.y}\ny_pred:\t{self.y_pred}\ny_sig:\t{self.y_sigma}\nstruc:\t{self.pdb_file}\n"

    __repr__ = __str__

    def initialize_user(self):
        """
        initializing a new library.
        """
        print(f"Initializing user '{self.user}'...")

        # handle app case
        if self.fname:
            fname = self.fname.split(".")[0]
            file_extension = self.fname.split(".")[-1]  # noqa: F841
        else:
            f = self.source.split("/")[-1]
            fname = f.split(".")[0]
            file_extension = f.split(".")[-1]  # noqa: F841

        # set paths
        self.source_path = os.path.join(USR_PATH, self.user, fname)
        self.rep_path = os.path.join(self.source_path, "zero_shot/rep")
        self.struc_path = os.path.join(self.source_path, "zero_shot/struc")

        # create user library if user does not exist
        if not os.path.exists(self.user):
            os.makedirs(self.user)
            if self.fname and not os.path.exists(
                os.path.join(self.user, self.fname.split(".")[0])
            ):
                fname = self.fname.split(".")[0]
                os.makedirs(os.path.join(self.user, f"{fname}"))
                os.makedirs(os.path.join(self.user, f"{fname}/library"))
                os.makedirs(os.path.join(self.user, f"{fname}/zero_shot"))
                os.makedirs(os.path.join(self.user, f"{fname}/design"))
            print(f"User created at {self.user}")

    def init_from_inheritance(self):
        pass

    # TODO: add protein specific loading functions of previously computed data (currently only handled in app)

    def init_from_file(self):

        # parse information
        if self.fname:
            fname = self.fname
        else:
            fname = self.source.split("/")[-1]

        # Check for representations
        if self.rep_path is None:
            fname = fname.split(".")[0]
            zs_path = os.path.join(self.user, f"{fname}/zero_shot")
            rep_path = os.path.join(zs_path, "rep")
            struc_path = os.path.join(zs_path, "struc")
            self.zs_path = zs_path
            self.rep_path = rep_path
            self.struc_path = struc_path
            self.name = fname
        else:
            rep_path = "/".join(self.rep_path.split("/")[:-1])

        # If file is not None, then initialize load fasta
        if self.source.endswith(".fasta"):
            self.load_fasta(file=self.source)

        # If file is not None, then initialize load fasta
        if self.source.endswith(".pdb"):
            self.load_structure(self.source, self.name)

    def load_fasta(self, file: str):
        """
        Load protein sequence from a FASTA file.

        Args:
            file (str): Path to the FASTA file.
        """
        sequences = []
        headers = []
        header = None
        seq = []

        with open(file, "r") as f:
            for line in f:
                line = line.strip()
                if line.startswith(">"):
                    if header:
                        sequences.append("".join(seq))
                        seq = []
                    header = line[1:]
                    headers.append(header)
                else:
                    seq.append(line)

            if seq:
                sequences.append("".join(seq))

        if not headers:
            raise ValueError(f"The file {file} is not in FASTA format or is empty.")

        if len(headers) > 1:
            warnings.warn(
                "The provided FASTA file contains multiple entries. Using only the first entry."
            )

        # Set the header as self.name
        name = headers[0].strip()  # remove potential leading spaces

        # Store the first sequence in self.seq
        seq = sequences[0]

        # Create and return a new Protein instance
        if self.fname:
            fname = self.fname.split(".")[0]
            self.name = fname
        else:
            self.name = name

        self.seq = seq

    def load_structure(self, prot_f, name=None, filter_solvent=True):
        """
        Load a structure from a pdb or cif file or an AtomArray.

        Args:
            struc: path to structure file or AtomArray
            name: provide protein name, else name of the file is assumed.
            filter_solvent (bool): True
        """
        if name is None and isinstance(prot_f, str):
            name = prot_f.split("/")[-1].split(".")[0]

        prot = struc.load_struc(prot_f)

        if filter_solvent:
            non_solvent_mask = biotite.structure.filter_solvent(prot)
            prot = prot[non_solvent_mask]

        seqs = struc.get_sequences(prot_f)
        chains = struc.chain_parser(prot_f)

        self.pdb_file = prot_f
        self.seq = seqs
        self.name = name
        self.struc = prot
        self.chains = chains

    def view_struc(self, color=None, highlight=None, sticks=None):
        """
        3D visualization of protein structure.

        Args:
            color (str): Choose different coloration options
        """
        view = struc.show_pdb(
            self.pdb_file, color=color, highlight=highlight, sticks=sticks
        )
        return view

    ### Zero-shot prediction ###
    def zs_prediction(
        self, model="esm2", batch_size=100, pbar=None, device=None, chain=None
    ):
        """
        Compute zero-shot scores

        Args:
            model (str): Model used to compute ZS scores
            batch_size (int): Batch size used to compute ZS-Scores
            pbar: App progress bar
            device (str): Choose hardware for computation. Default 'None' for autoselection
                        other options are 'cpu' and 'cuda'.
        """

        # Set a default chain if none is provided and there are chains available
        if chain is None and len(self.chains) >= 1:
            chain = self.chains[0]

        # Now ensure chain has a value before proceeding
        if chain is not None and len(self.chains) >= 1:
            seq = self.seq[chain]
            dest = os.path.join(self.zs_path, "results", chain, model)
        else:
            seq = self.seq
            dest = os.path.join(self.zs_path, "results", model)

        # Check if results already exist
        if os.path.exists(dest):
            print(f"Results already computed. Loading from {dest}")
            p = torch.load(os.path.join(dest, "prob_dist.pt"))
            mmp = torch.load(os.path.join(dest, "masked_marginal_probability.pt"))
            entropy = torch.load(os.path.join(dest, "per_position_entropy.pt"))
            logits = torch.load(os.path.join(dest, "masked_logits.pt"))
            df = pd.read_csv(os.path.join(dest, "zs_scores.csv"))
        else:
            # Perform computation if results do not exist
            print("Computing logits")
            logits, alphabet = esm_tools.get_mutant_logits(
                seq, batch_size=batch_size, model=model, pbar=pbar, device=device
            )

            # Calculations
            p = esm_tools.get_probability_distribution(logits)
            mmp = esm_tools.masked_marginal_probability(p, seq, alphabet)
            entropy = esm_tools.per_position_entropy(p)

            # Create directory if it doesn't exist
            if not os.path.exists(dest):
                os.makedirs(dest)
                print(f"Directory created at {dest}")

            # Save tensors
            torch.save(p, os.path.join(dest, "prob_dist.pt"))
            torch.save(mmp, os.path.join(dest, "masked_marginal_probability.pt"))
            torch.save(entropy, os.path.join(dest, "per_position_entropy.pt"))
            torch.save(logits, os.path.join(dest, "masked_logits.pt"))

            self.p = p
            self.mmp = mmp
            self.entropy = entropy
            self.logits = logits

            df = esm_tools.zs_to_csv(
                seq, alphabet, p, mmp, entropy, os.path.join(dest, "zs_scores.csv")
            )

            # no true y_values
            ys = [None] * len(mmp)  # noqa: F841

        out = {
            "df": df,
            "rep_path": self.rep_path,
            "struc_path": self.struc_path,
            "y_type": "num",
            "y_pred_col": "mmp",
            "seqs_col": "sequence",
            "names_col": "mutant",
            "reps": self.reps,
            "class_dict": self.class_dict,
            "pred_data": True,
        }

        return out

    def zs_library(self, model="esm2", chain=None):
        """
        Generate zero-shot library.
        """

        if chain is None and len(self.chains) >= 1:
            chain = self.chains[0]
            wt_seq = self.seq[chain]
            zs_results_path = os.path.join(
                self.zs_path, "results", chain, model, "zs_scores.csv"
            )
        elif len(self.chains) >= 1:
            wt_seq = self.seq[chain]
            zs_results_path = os.path.join(
                self.zs_path, "results", chain, model, "zs_scores.csv"
            )
        else:
            wt_seq = self.seq
            zs_results_path = os.path.join(
                self.zs_path, "results", model, "zs_scores.csv"
            )

        # load already computed zs scores
        if os.path.exists(zs_results_path):
            df = pd.read_csv(zs_results_path)

        # generate df with blank y-values
        else:
            # wt_seq = self.seq
            canonical_aas = [
                "A",
                "R",
                "N",
                "D",
                "C",
                "Q",
                "E",
                "G",
                "H",
                "I",
                "L",
                "K",
                "M",
                "F",
                "P",
                "S",
                "T",
                "W",
                "Y",
                "V",
            ]
            mutants, sequences = [], []
            for pos in range(len(wt_seq)):
                for aa in canonical_aas:
                    if wt_seq[pos] != aa:
                        mutants.append(wt_seq[pos] + str(pos + 1) + aa)
                        sequences.append(wt_seq[:pos] + aa + wt_seq[pos + 1 :])
            ys = [None] * len(mutants)

            df = pd.DataFrame(
                {
                    "mutant": mutants,
                    "sequence": sequences,
                    "p": ys,
                    "mmp": ys,
                    "entropy": ys,
                }
            )

        out = {
            "df": df,
            "rep_path": self.rep_path,
            "struc_path": self.struc_path,
            "y_type": "num",
            "y_col": "mmp",
            "seqs_col": "sequence",
            "names_col": "mutant",
            "reps": self.reps,
            "class_dict": self.class_dict,
        }

        return out

    ### Structure prediction ###
    def esm_fold(
        self, batch_size=100, chain=None, dest=None, pbar=None
    ):  # If structure prediction will become available for libraries, set dest in library, create a protein dir under the file name
        """
        Compute zero-shot scores

        Args:
            batch_size (int): Batch size used to compute ZS-Scores
            dest (str): custom destination for file
            pbar: App progress bar

        """
        if chain is None:
            chain = self.chains[0]

        # check scores for this protein and model have already been computed

        user_path = self.user

        if self.name and not dest:
            dest = os.path.join(user_path, "protein/")
            pdb_file = os.path.join(dest, {self.name})
        elif dest:
            pdb_file = os.path.join(dest, f"{self.name}.pdb")

        # Check if structure already exist
        if os.path.exists(pdb_file):
            self.pdb_file = pdb_file
            self.struc = strucio.load_structure(pdb_file)
            self.chains = struc.chain_parser(self.struc)
            # get ptsm and plddts from loaded files
        else:
            os.makedirs(dest, exist_ok=True)
            all_headers, all_sequences, all_pdbs, pTMs, mean_pLDDTs = (
                esm_tools.structure_prediction(seqs=[self.seq], names=[self.name])
            )
            pdb = all_pdbs[0]
            pdb.write(pdb_file)

            self.pdb_file = pdb_file
            self.struc = all_pdbs[0]
            self.pTMs = pTMs[0]
            self.pLDDT = mean_pLDDTs[0]
            self.chains = struc.chain_parser(self.struc)

        return self.struc

    ### Inverse Folding ###
    def esm_if(
        self,
        fixed=[],
        target_chain=None,
        chains=None,
        temperature=1.0,
        num_samples=100,
        model=None,
        pbar=None,
        dest=None,
        noise=0.2,
    ):
        """
        Perform inverse folding using ESM-IF for multi-chain structures.

        Args:
            fixed (list): List of fixed residues (numbers starting from 1).
            target_chain (str): string that should be designed.
            chains (list): chains that should be loaded from the complex.
            temperatrue (float): sampling temperature used.
            num_samples (int): number of samples generated.
            model (str): Type of model used.
            pbar: Progress bar used by shiny app.
            dest (str): Custom save destination.
            noise (float): Noise added to backbone structure.
        """
        user_path = self.user

        if dest:
            csv_path = os.path.join(dest, f"{self.name}.csv")
        else:
            dest = os.path.join(user_path, f"{self.name}/design/")
            csv_path = os.path.join(dest, f"{self.name}.csv")

        os.makedirs(dest, exist_ok=True)

        # define chains
        if chains is None:
            chains = self.chains

        if target_chain is None:
            target_chain = chains[0]
            assert target_chain in chains, "Target chain must be in chains."

        # return dataframe of results
        df = esm_tools.esm_design(
            self.pdb_file,
            target_chain,
            chains,
            fixed=fixed,
            temperature=temperature,
            num_samples=num_samples,
            model=model,
            alphabet=esm_tools.alphabet,
            noise=noise,
            pbar=pbar,
        )

        df.to_csv(csv_path)

        self.designs = csv_path

        rep_path = os.path.join(dest, "rep")
        struc_path = os.path.join(dest, "struc")

        out = {
            "df": df,
            "rep_path": rep_path,
            "struc_path": struc_path,
            "y_type": "num",
            "y_col": "log_likelihood",
            "seqs_col": "sequence",
            "names_col": "names",
            "reps": [],
            "class_dict": self.class_dict,
        }

        return out

    # Plot
    # Plot zero-shot entropy
    def plot_entropy(self, model="esm2", title=None, section=None, chain=None):
        if chain is None and len(self.chains) >= 1:
            chain = self.chains[0]
            seq = self.seq[chain]
        elif len(self.chains) >= 1:
            seq = self.seq[chain]
        else:
            seq = self.seq
            chain = None

        if self.name:
            dest = os.path.join(self.user, f"{self.name}/zero_shot/results", model)
            if chain:
                dest = os.path.join(
                    self.user, f"{self.name}/zero_shot/results/{chain}", model
                )
        else:
            dest = os.path.join(self.user, "protein/zero_shot/results", model)
            if chain:
                dest = os.path.join(
                    self.user, f"protein/zero_shot/results/{chain}", model
                )

        # Load required data
        self.p = torch.load(os.path.join(dest, "prob_dist.pt"))
        self.mmp = torch.load(os.path.join(dest, "masked_marginal_probability.pt"))
        self.entropy = torch.load(os.path.join(dest, "per_position_entropy.pt"))
        self.logits = torch.load(os.path.join(dest, "masked_logits.pt"))

        # Section handling
        seq_len = len(seq)
        if section is None:
            section = (0, seq_len)
        elif isinstance(section, tuple):
            if len(section) != 2 or any(not isinstance(i, int) for i in section):
                raise ValueError("Section must be a tuple of two integers.")
            if section[0] < 0 or section[0] >= seq_len or section[1] > seq_len:
                raise ValueError("Section indices are out of sequence range.")
            if section[1] < section[0]:
                raise ValueError("Section start index must be less than end index.")
        else:
            raise TypeError("Section must be a tuple or None.")

        # Set title
        if title is None:
            title = f"{model_dict[model]} per-position entropy"

        # Plot entropy
        fig = esm_tools.plot_per_position_entropy(
            per_position_entropy=self.entropy,
            sequence=seq,
            highlight_positions=None,
            dest=None,
            title=title,
            section=section,
        )
        return fig

    def plot_scores(
        self,
        model="esm2",
        section=None,
        color_scheme=None,
        title=None,
        highlight_positions=None,
        chain=None,
    ):
        """
        Plot the zero-shot prediction scores for a given model and sequence.

        Args:
            model (str): The name of the model to use for plotting (default: 'esm2').
            section (tuple): Section of the sequence to be shown in the plot - low and high end of sequence to be displayed. Show entire sequence if None (default: None).
            color_scheme (str): Color scheme for the heatmap ('rwb' for red-white-blue, 'r' for reds, 'b' for blues) (default: None).
            title (str): Title of the plot (default: None).
            highlight_positions (dict): Dictionary specifying positions to highlight with the format {position: residue} (default: None).

        Returns:
            fig (matplotlib.figure.Figure): The created matplotlib figure.
        """

        if chain is None and len(self.chains) >= 1:
            chain = self.chains[0]
            seq = self.seq[chain]
        elif len(self.chains) >= 1:
            seq = self.seq[chain]
        else:
            seq = self.seq
            chain = None

        if self.name:
            dest = os.path.join(self.user, f"{self.name}/zero_shot/results", model)
            if chain:
                dest = os.path.join(
                    self.user, f"{self.name}/zero_shot/results/{chain}", model
                )
        else:
            dest = os.path.join(self.user, "protein/zero_shot/results", model)
            if chain:
                dest = os.path.join(
                    self.user, f"protein/zero_shot/results/{chain}", model
                )

        # Load required data
        self.p = torch.load(os.path.join(dest, "prob_dist.pt"))
        self.mmp = torch.load(os.path.join(dest, "masked_marginal_probability.pt"))
        self.entropy = torch.load(os.path.join(dest, "per_position_entropy.pt"))
        self.logits = torch.load(os.path.join(dest, "masked_logits.pt"))

        # Section handling
        seq_len = len(seq)
        if section is None:
            section = (0, seq_len)
        elif isinstance(section, tuple):
            if len(section) != 2 or any(not isinstance(i, int) for i in section):
                raise ValueError("Section must be a tuple of two integers.")
            if section[0] < 0 or section[0] >= seq_len or section[1] > seq_len:
                raise ValueError("Section indices are out of sequence range.")
            if section[1] < section[0]:
                raise ValueError("Section start index must be less than end index.")
        else:
            raise TypeError("Section must be a tuple or None.")

        # Set color scheme
        if color_scheme is None:
            color_scheme = "rwb"

        # Set title
        if title is None:
            title = f"{model_dict[model]} Zero-shot prediction scores"

        # Plot heatmap
        fig = esm_tools.plot_heatmap(
            p=self.mmp,
            alphabet=esm_tools.alphabet,
            dest=None,
            title=title,
            show=False,
            remove_tokens=True,
            color_sheme=color_scheme,
            section=section,
            highlight_positions=highlight_positions,
        )
        return fig

    # structure utils
    def get_contacts(
        self, chain: Union[str, None] = None, target: str = "protein", dist=7.0
    ):
        """
        Get protein protein contacts for a specific chain in a protein.

        Args:
            chain (str): specify chain for which to compute the contacts. Default 'None' will take the first chain.
            target (str): Specify protein-'protein' contacts or protein-'ligand' contacts, Default 'protein'
        """
        return struc.get_contacts(self.struc, chain, target, dist)

    ### getters and setters ###
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError(
                f"Expected 'name' to be of type 'str', but got '{type(value).__name__}'"
            )
        self._name = value

    # For seq
    @property
    def seq(self):
        return self._seq

    @seq.setter
    def seq(self, value):
        if not isinstance(value, (str, dict)) and value is not None:
            raise TypeError(
                f"Expected 'seq' to be of type 'str', but got '{type(value).__name__}'"
            )
        self._seq = value

    # For rep
    @property
    def reps(self):
        return self._reps

    @reps.setter
    def reps(self, value):
        if not isinstance(value, (list, tuple)) and value is not None:
            raise TypeError(
                f"Expected 'rep' to be of type 'list' or 'tuple', but got '{type(value).__name__}'"
            )
        self._reps = list(value)

    # For y
    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        if not isinstance(value, (str, int, float)) and value is not None:
            raise TypeError(
                f"Expected 'rep' to be of type 'int', 'float', or 'str', but got '{type(value).__name__}'"
            )
        self._y = value



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

"""
A subpackage for mining_tools.
"""

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from proteusAI.io_tools.embeddings import *  # noqa: F403
from proteusAI.io_tools.fasta import *  # noqa: F403



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from tempfile import gettempdir
import biotite.sequence.io.fasta as fasta
import biotite.application.blast as blast
import biotite.database.entrez as entrez


# TODO: change email
def search_related_sequences(
    query: str,
    program: str = "blastp",
    database: str = "nr",
    obey_rules: bool = True,
    mail: str = "johnyfunk@gmail.com",
):
    """
    Search for related sequences using the plast web app.

    Parameters:
        query (str):
            Query sequence
        program (str, optional):
            The specific BLAST program. One of 'blastn', 'megablast', 'blastp', 'blastx', 'tblastn' and 'tblastx'.
        database (str, optional):
            The NCBI sequence database to blast against. By default it contains all sequences (`database`='nr'`).
        obey_rules (bool, optional):
            If true, the application raises an :class:`RuleViolationError`, if the server is contacted too often,
            based on the NCBI BLAST usage rules. (Default: True)
        mail : str, optional
            If a mail address is provided, it will be appended in the
            HTTP request. This allows the NCBI to contact you in case
            your application sends too many requests.

    Returns:
        tuple: two lits, the first containing the hits and the second the hit sequences

    Example:
        hits, hit_seqs = blast_related_sequences(query=sequence, database='swissprot')
    """
    # Search only the UniProt/SwissProt database
    blast_app = blast.BlastWebApp(
        program=program,
        query=query,
        database=database,
        obey_rules=obey_rules,
        mail=mail,
    )
    blast_app.start()
    blast_app.join()
    alignments = blast_app.get_alignments()
    # Get hit IDs for hits with score > 200
    hits = []
    for ali in alignments:
        if ali.score > 200:
            hits.append(ali.hit_id)
    # Get the sequences from hit IDs
    hit_seqs = []
    for hit in hits:
        file_name = entrez.fetch(hit, gettempdir(), "fa", "protein", "fasta")
        fasta_file = fasta.FastaFile.read(file_name)
        hit_seqs.append(fasta.get_sequence(fasta_file))

    return hits, hit_seqs



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

"""
A subpackage for pytorch tools.
"""

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from proteusAI.ml_tools.torch_tools.torch_tools import *  # noqa: F403



import matplotlib.pyplot as plt
import seaborn as sns
from typing import Union, List
import pandas as pd
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
import numpy as np
import umap

representation_dict = {
    "One-hot": "ohe",
    "BLOSUM50": "blosum50",
    "BLOSUM62": "blosum62",
    "ESM-2": "esm2",
    "ESM-1v": "esm1v",
}


def plot_predictions_vs_groundtruth(
    y_true: list,
    y_pred: list,
    title: Union[str, None] = None,
    x_label: Union[str, None] = None,
    y_label: Union[str, None] = None,
    plot_grid: bool = True,
    file: Union[str, None] = None,
    show_plot: bool = True,
):
    # Create the plot
    fig, ax = plt.subplots(figsize=(10, 5))
    sns.scatterplot(x=y_true, y=y_pred, alpha=0.5, ax=ax)

    # Set plot title and labels
    if title is None:
        title = "Predicted vs. True y-values"
    if y_label is None:
        y_label = "predicted y"
    if x_label is None:
        x_label = "y"

    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.plot(
        [min(y_true), max(y_true)],
        [min(y_true), max(y_true)],
        color="grey",
        linestyle="dotted",
        linewidth=2,
    )  # diagonal line
    ax.grid(plot_grid)

    # Save the plot to a file
    if file is not None:
        fig.savefig(file)

    # Show the plot
    if show_plot:
        plt.show()

    # Return the figure and axes
    return fig, ax


def plot_tsne(
    x: List[np.ndarray],
    y: Union[List[Union[float, str]], None] = None,
    y_upper: Union[float, None] = None,
    y_lower: Union[float, None] = None,
    names: Union[List[str], None] = None,
    y_type: str = "num",
    random_state: int = 42,
    rep_type: Union[str, None] = None,
    highlight_mask: Union[List[Union[int, float]], None] = None,
    highlight_label: str = "Highlighted",
):
    """
    Create a t-SNE plot and optionally color by y values, with special coloring for points outside given thresholds.
    Handles cases where y is None or a list of Nones by not applying hue.
    Optionally highlights points based on the highlight_mask.

    Args:
        x (List[np.ndarray]): List of sequence representations as numpy arrays.
        y (List[Union[float, str]]): List of y values, can be None or contain None.
        y_upper (float): Upper threshold for special coloring.
        y_lower (float): Lower threshold for special coloring.
        names (List[str]): List of names for each point.
        y_type (str): 'class' for categorical labels or 'num' for numerical labels.
        random_state (int): Random state.
        rep_type (str): Representation type used for plotting.
        highlight_mask (List[Union[int, float]]): List of mask values, non-zero points will be highlighted.
        highlight_label (str): Text for the legend entry of highlighted points.
    """
    fig, ax = plt.subplots(figsize=(10, 5))

    x = np.array([t.numpy() if hasattr(t, "numpy") else t for t in x])

    if len(x.shape) == 3:  # Flatten if necessary
        x = x.reshape(x.shape[0], -1)

    tsne = TSNE(n_components=2, verbose=1, random_state=random_state)
    z = tsne.fit_transform(x)

    df = pd.DataFrame(z, columns=["z1", "z2"])
    df["y"] = y if y is not None and any(y) else None  # Use y if it's informative
    if names and len(names) == len(y):
        df["names"] = names
    else:
        df["names"] = [None] * len(y)

    # Handle the palette based on whether y is numerical or categorical
    if isinstance(y[0], (int, float)):  # If y is numerical
        cmap = sns.cubehelix_palette(rot=-0.2, as_cmap=True)
    else:  # If y is categorical
        cmap = sns.color_palette("Set2", as_cmap=False)

    hue = (
        "y" if df["y"].isnull().sum() != len(df["y"]) else None
    )  # Use hue only if y is informative
    scatter = sns.scatterplot(
        x="z1", y="z2", hue=hue, palette=cmap if hue else None, data=df
    )

    # Apply special coloring only if y values are valid and thresholds are provided
    if hue and (y_upper is not None or y_lower is not None):
        outlier_mask = (
            (df["y"] > y_upper)
            if y_upper is not None
            else np.zeros(len(df), dtype=bool)
        )
        outlier_mask |= (
            (df["y"] < y_lower)
            if y_lower is not None
            else np.zeros(len(df), dtype=bool)
        )
        scatter.scatter(
            df["z1"][outlier_mask], df["z2"][outlier_mask], color="lightgrey"
        )

    # Highlight points based on the highlight_mask
    if highlight_mask is not None:
        highlight_mask = np.array(highlight_mask)
        highlight_points = highlight_mask != 0  # Non-zero entries in the highlight_mask
        scatter.scatter(
            df["z1"][highlight_points],
            df["z2"][highlight_points],
            color="red",
            marker="x",
            s=60,
            alpha=0.7,
            label=highlight_label,
        )

    scatter.set_title(f"t-SNE projection of {rep_type if rep_type else 'data'}")

    # Add the legend, making sure to include highlighted points
    handles, labels = scatter.get_legend_handles_labels()
    if highlight_label in labels:
        ax.legend(handles, labels, title="Legend")
    else:
        ax.legend(title="Legend")

    return fig, ax, df


def plot_umap(
    x: List[np.ndarray],
    y: Union[List[Union[float, str]], None] = None,
    y_upper: Union[float, None] = None,
    y_lower: Union[float, None] = None,
    names: Union[List[str], None] = None,
    y_type: str = "num",
    random_state: int = 42,
    rep_type: Union[str, None] = None,
    highlight_mask: Union[List[Union[int, float]], None] = None,
    highlight_label: str = "Highlighted",
):
    """
    Create a UMAP plot and optionally color by y values, with special coloring for points outside given thresholds.
    Handles cases where y is None or a list of Nones by not applying hue.
    Optionally highlights points based on the highlight_mask.

    Args:
        x (List[np.ndarray]): List of sequence representations as numpy arrays.
        y (List[Union[float, str]]): List of y values, can be None or contain None.
        y_upper (float): Upper threshold for special coloring.
        y_lower (float): Lower threshold for special coloring.
        names (List[str]): List of names for each point.
        y_type (str): 'class' for categorical labels or 'num' for numerical labels.
        random_state (int): Random state.
        rep_type (str): Representation type used for plotting.
        highlight_mask (List[Union[int, float]]): List of mask values, non-zero points will be highlighted.
        highlight_label (str): Text for the legend entry of highlighted points.
    """
    fig, ax = plt.subplots(figsize=(10, 5))

    x = np.array([t.numpy() if hasattr(t, "numpy") else t for t in x])

    if len(x.shape) == 3:  # Flatten if necessary
        x = x.reshape(x.shape[0], -1)

    umap_model = umap.UMAP(n_components=2, random_state=random_state)
    z = umap_model.fit_transform(x)

    df = pd.DataFrame(z, columns=["z1", "z2"])
    df["y"] = y if y is not None and any(y) else None  # Use y if it's informative
    if names and len(names) is not None and len(names) == len(y):
        df["names"] = names
    else:
        df["names"] = [None] * len(y)

    # Handle the palette based on whether y is numerical or categorical
    if isinstance(y[0], (int, float)):  # If y is numerical
        cmap = sns.cubehelix_palette(rot=-0.2, as_cmap=True)
    else:  # If y is categorical
        cmap = sns.color_palette("Set2", as_cmap=False)

    hue = (
        "y" if df["y"].isnull().sum() != len(df["y"]) else None
    )  # Use hue only if y is informative
    scatter = sns.scatterplot(
        x="z1", y="z2", hue=hue, palette=cmap if hue else None, data=df
    )

    # Apply special coloring only if y values are valid and thresholds are provided
    if hue and (y_upper is not None or y_lower is not None):
        outlier_mask = (
            (df["y"] > y_upper)
            if y_upper is not None
            else np.zeros(len(df), dtype=bool)
        )
        outlier_mask |= (
            (df["y"] < y_lower)
            if y_lower is not None
            else np.zeros(len(df), dtype=bool)
        )
        scatter.scatter(
            df["z1"][outlier_mask], df["z2"][outlier_mask], color="lightgrey"
        )

    # Highlight points based on the highlight_mask
    if highlight_mask is not None:
        highlight_mask = np.array(highlight_mask)
        highlight_points = highlight_mask != 0  # Non-zero entries in the highlight_mask
        scatter.scatter(
            df["z1"][highlight_points],
            df["z2"][highlight_points],
            color="red",
            marker="x",
            s=60,
            alpha=0.7,
            label=highlight_label,
        )

    scatter.set_title(f"UMAP projection of {rep_type if rep_type else 'data'}")

    # Add the legend, making sure to include highlighted points
    handles, labels = scatter.get_legend_handles_labels()
    if highlight_label in labels:
        ax.legend(handles, labels, title="Legend")
    else:
        ax.legend(title="Legend")

    return fig, ax, df


def plot_pca(
    x: List[np.ndarray],
    y: Union[List[Union[float, str]], None] = None,
    y_upper: Union[float, None] = None,
    y_lower: Union[float, None] = None,
    names: Union[List[str], None] = None,
    y_type: str = "num",
    random_state: int = 42,
    rep_type: Union[str, None] = None,
    highlight_mask: Union[List[Union[int, float]], None] = None,
    highlight_label: str = "Highlighted",
):
    """
    Create a PCA plot and optionally color by y values, with special coloring for points outside given thresholds.
    Handles cases where y is None or a list of Nones by not applying hue.
    Optionally highlights points based on the highlight_mask.

    Args:
        x (List[np.ndarray]): List of sequence representations as numpy arrays.
        y (List[Union[float, str]]): List of y values, can be None or contain None.
        y_upper (float): Upper threshold for special coloring.
        y_lower (float): Lower threshold for special coloring.
        names (List[str]): List of names for each point.
        y_type (str): 'class' for categorical labels or 'num' for numerical labels.
        random_state (int): Random state.
        rep_type (str): Representation type used for plotting.
        highlight_mask (List[Union[int, float]]): List of mask values, non-zero points will be highlighted.
        highlight_label (str): Text for the legend entry of highlighted points.
    """
    fig, ax = plt.subplots(figsize=(10, 5))

    x = np.array([t.numpy() if hasattr(t, "numpy") else t for t in x])

    if len(x.shape) == 3:  # Flatten if necessary
        x = x.reshape(x.shape[0], -1)

    pca = PCA(n_components=2, random_state=random_state)
    z = pca.fit_transform(x)

    df = pd.DataFrame(z, columns=["z1", "z2"])
    df["y"] = y if y is not None and any(y) else None  # Use y if it's informative
    if names and len(names) == len(y):
        df["names"] = names
    else:
        df["names"] = [None] * len(y)

    # Handle the palette based on whether y is numerical or categorical
    if isinstance(y[0], (int, float)):  # If y is numerical
        cmap = sns.cubehelix_palette(rot=-0.2, as_cmap=True)
    else:  # If y is categorical
        cmap = sns.color_palette("Set2", as_cmap=False)

    hue = (
        "y" if df["y"].isnull().sum() != len(df["y"]) else None
    )  # Use hue only if y is informative
    scatter = sns.scatterplot(
        x="z1", y="z2", hue=hue, palette=cmap if hue else None, data=df
    )

    # Apply special coloring only if y values are valid and thresholds are provided
    if hue and (y_upper is not None or y_lower is not None):
        outlier_mask = (
            (df["y"] > y_upper)
            if y_upper is not None
            else np.zeros(len(df), dtype=bool)
        )
        outlier_mask |= (
            (df["y"] < y_lower)
            if y_lower is not None
            else np.zeros(len(df), dtype=bool)
        )
        scatter.scatter(
            df["z1"][outlier_mask], df["z2"][outlier_mask], color="lightgrey"
        )

    # Highlight points based on the highlight_mask
    if highlight_mask is not None:
        highlight_mask = np.array(highlight_mask)
        highlight_points = highlight_mask != 0  # Non-zero entries in the highlight_mask
        scatter.scatter(
            df["z1"][highlight_points],
            df["z2"][highlight_points],
            color="red",
            marker="x",
            s=60,
            alpha=0.7,
            label=highlight_label,
        )

    scatter.set_title(f"PCA projection of {rep_type if rep_type else 'data'}")

    # Add the legend, making sure to include highlighted points
    handles, labels = scatter.get_legend_handles_labels()
    if highlight_label in labels:
        ax.legend(handles, labels, title="Legend")
    else:
        ax.legend(title="Legend")

    return fig, ax, df



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from biotite.structure.io.mol import MOLFile
from biotite.structure.io.pdb import PDBFile
import numpy as np
import biotite.structure as struc
import biotite.structure.io as strucio
from Bio import SeqIO
import py3Dmol
from string import ascii_uppercase, ascii_lowercase


alphabet_list = list(ascii_uppercase + ascii_lowercase)
pymol_color_list = [
    "#33ff33",
    "#00ffff",
    "#ff33cc",
    "#ffff00",
    "#ff9999",
    "#e5e5e5",
    "#7f7fff",
    "#ff7f00",
    "#7fff7f",
    "#199999",
    "#ff007f",
    "#ffdd5e",
    "#8c3f99",
    "#b2b2b2",
    "#007fff",
    "#c4b200",
    "#8cb266",
    "#00bfbf",
    "#b27f7f",
    "#fcd1a5",
    "#ff7f7f",
    "#ffbfdd",
    "#7fffff",
    "#ffff7f",
    "#00ff7f",
    "#337fcc",
    "#d8337f",
    "#bfff3f",
    "#ff7fff",
    "#d8d8ff",
    "#3fffbf",
    "#b78c4c",
    "#339933",
    "#66b2b2",
    "#ba8c84",
    "#84bf00",
    "#b24c66",
    "#7f7f7f",
    "#3f3fa5",
    "#a5512b",
]


def get_atom_array(file_path):
    """
    Returns atom array for a pdb file.

    Parameters:
        file_path (str): path to pdb file

    Returns:
        atom array
    """
    if file_path.endswith("pdb"):
        atom_mol = PDBFile.read(file_path)
        atom_array = atom_mol.get_structure()
    else:
        try:
            atom_mol = MOLFile.read(file_path)
            atom_array = atom_mol.get_structure()
        except Exception as e:
            raise ValueError(f"File: {file_path} has an invalid format. Error: {e}")

    return atom_array


def mol_contacts(mols, protein, dist=4.0):
    """
    Get residue ids of contacts of small molecule(s) to protein.

    Parameters:
        mols (str, list of strings): path to molecule file
        protein (str): path to protein file
        dist (float): distance to be considered contact

    Returns:
        set: res_ids of protein residues which are in contact with molecules
    """
    if isinstance(mols, list) or isinstance(mols, tuple):
        mols = [get_atom_array(m) for m in mols]
    else:
        mols = [get_atom_array(mols)]

    protein = get_atom_array(protein)

    res_ids = set()
    for mol in mols:
        cell_list = struc.CellList(mol, cell_size=dist)
        for prot in protein:
            contacts = cell_list.get_atoms(prot.coord, radius=dist)

            contact_indices = np.where((contacts != -1).any(axis=1))[0]

            contact_res_ids = prot.res_id[contact_indices]
            res_ids.update(contact_res_ids)

    res_ids = sorted(res_ids)
    return res_ids


def pdb_to_fasta(pdb_path: str):
    """
    Returns fasta sequence of pdb file.

    Parameters:
        pdb_path (str): path to pdb file

    Returns:
        str: fasta sequence string
    """
    seq = ""
    with open(pdb_path, "r") as pdb_file:
        for record in SeqIO.parse(pdb_file, "pdb-atom"):
            seq = "".join([">", str(record.id), "\n", str(record.seq)])

    return seq


def get_sse(pdb_path: str):
    """
    Returns the secondary structure of a protein given the pdb file.
    The secondary structure is infered using the P-SEA algorithm
    as imple-mented by the biotite Python package.

    Parameters:
        pdb_path (str): path to pdb
    """
    array = strucio.load_structure(pdb_path)
    sse = struc.annotate_sse(array)
    sse = "".join(sse)
    return sse


### Visualization
def show_pdb(
    pdb_path,
    color="confidence",
    vmin=50,
    vmax=90,
    chains=None,
    Ls=None,
    size=(800, 480),
    show_sidechains=False,
    show_mainchains=False,
    highlight=None,
):
    """
    This function displays the 3D structure of a protein from a given PDB file in a Jupyter notebook.
    The protein structure can be colored by chain, rainbow, pLDDT, or confidence value. The size of the
    display can be changed. The sidechains and mainchains can be displayed or hidden.
    Parameters:
        pdb_path (str): The filename of the PDB file that contains the protein structure.
        color (str, optional): The color scheme for the protein structure. Can be "chain", "rainbow", "pLDDT", or "confidence". Defaults to "rainbow".
        vmin (float, optional): The minimum value of pLDDT or confidence value. Defaults to 50.
        vmax (float, optional): The maximum value of pLDDT or confidence value. Defaults to 90.
        chains (int, optional): The number of chains to be displayed. Defaults to None.
        Ls (list, optional): A list of the chains to be displayed. Defaults to None.
        size (tuple, optional): The size of the display window. Defaults to (800, 480).
        show_sidechains (bool, optional): Whether to display the sidechains. Defaults to False.
        show_mainchains (bool, optional): Whether to display the mainchains. Defaults to False.
    Returns:
        view: The 3Dmol view object that displays the protein structure.
    """

    with open(pdb_path) as ifile:
        system = "".join([x for x in ifile])

    view = py3Dmol.view(
        js="https://3dmol.org/build/3Dmol.js", width=size[0], height=size[1]
    )

    if chains is None:
        chains = 1 if Ls is None else len(Ls)

    view.addModelsAsFrames(system)
    if color == "pLDDT" or color == "confidence":
        view.setStyle(
            {
                "cartoon": {
                    "colorscheme": {
                        "prop": "b",
                        "gradient": "rwb",
                        "min": vmin,
                        "max": vmax,
                    }
                }
            }
        )
    elif color == "rainbow":
        view.setStyle({"cartoon": {"color": "spectrum"}})
    elif color == "chain":
        for n, chain, color in zip(range(chains), alphabet_list, pymol_color_list):
            view.setStyle({"chain": chain}, {"cartoon": {"color": color}})

    if show_sidechains:
        BB = ["C", "O", "N"]
        view.addStyle(
            {
                "and": [
                    {"resn": ["GLY", "PRO"], "invert": True},
                    {"atom": BB, "invert": True},
                ]
            },
            {"stick": {"colorscheme": "WhiteCarbon", "radius": 0.3}},
        )
        view.addStyle(
            {"and": [{"resn": "GLY"}, {"atom": "CA"}]},
            {"sphere": {"colorscheme": "WhiteCarbon", "radius": 0.3}},
        )
        view.addStyle(
            {"and": [{"resn": "PRO"}, {"atom": ["C", "O"], "invert": True}]},
            {"stick": {"colorscheme": "WhiteCarbon", "radius": 0.3}},
        )
    if show_mainchains:
        BB = ["C", "O", "N", "CA"]
        view.addStyle(
            {"atom": BB}, {"stick": {"colorscheme": "WhiteCarbon", "radius": 0.3}}
        )
        view.zoomTo()

    view.zoomTo()

    return view



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

"""
A subpackage for sklearn_tools activity_prediction.
"""

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from proteusAI.visual_tools.plots import *  # noqa: F403



# This source code is part of the proteusAI package and is distributed
# under the MIT License.


from importlib import metadata

__version__ = metadata.version("proteusAI")
__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from .Protein import *  # noqa: F403
from .Library import *  # noqa: F403
from .Model import *  # noqa: F403



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

"""
A subpackage is concerned with the Model object of ProteusAI.
"""

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from .model import Model  # noqa: F401



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

import os
import numpy as np
from proteusAI.design_tools import Constraints
import pandas as pd


class ZeroShot:
    """
    ZeroShot inference for single mutational effects. Mutates every single position to every possible amino acid at
    that position. Then calculate differnt metrics of structure and embeddings for the mutants. Observe which mutations
    have the greatest effect on the mutants.

    Parameters:
    -----------
        seq (str): native sequence to be optimized
        constraints (dict): constraints on sequence.
            Keys describe the kind of constraint and values the position on which they act.
        sampler (str): choose between simulated_annealing and substitution design_tools.
            Default 'simulated annealing'
        batch_size (int): number of independent trajectories per sampling step. Lowest energy mutant will be selected when
            multiple are viable. Default 16
        steps (int): number of sampling steps per trajectory.
            For simulated annealing, the number of iterations is often chosen in the range of [1,000, 10,000].
        T (float): sampling temperature.
            For simulated annealing, T0 is often chosen in the range [1, 100]. default 1
        M (float): rate of temperature decay.
            or simulated annealing, a is often chosen in the range [0.01, 0.1] or [0.001, 0.01]. Default 0.01
        mut_p (tuple): probabilities for substitution, insertion and deletion.
            Default [0.6, 0.2, 0.2]
        pred_struc (bool): if True predict the structure of the protein at every step and use structure
            based constraints in the energy function. Default True.
        max_len (int): maximum length sequence length for lenght constraint.
            Default 300.
        w_len (float): weight of length constraint.
            Default 0.01
        w_identity (float): Weight of sequence identity constraint. Positive values reward low sequence identity to native sequence.
            Default 0.04
        w_ptm (float): weight for ptm. pTM is calculated as 1-pTM, because lower energies should be better.
            Default 1.
        w_plddt (float): weight for plddt.  The mean pLDDT is calculated as 1-mean_pLDDT, because lower energies should be better.
            Default 1.
        w_globularity (float): weight of globularity constraint
            Default 0.001
        w_bb_coord (float): weight on backbone coordination constraint. Constraints backbone to native structure.
            Default 0.02
        w_all_atm (float): weight on all atom coordination constraint. Acts on all atoms which are constrained
            Default 0.15
        w_sasa (float): weight of surface exposed hydrophobics constraint
            Default 0.02
        outdir (str): path to output directory.
            Default None
        verbose (bool): if verbose print information
    """

    def __init__(
        self,
        seq: str = None,
        name="Prot",
        constraints=None,
        batch_size: int = 20,
        pred_struc: bool = True,
        w_ptm: float = 1,
        w_plddt: float = 1,
        w_globularity: float = 0.001,
        w_bb_coord: float = 0.02,
        w_all_atm: float = 0.15,
        w_sasa: float = 0.02,
        outdir: str = None,
        verbose: bool = False,
    ):

        if constraints is None:
            constraints = {"all_atm": []}
        self.seq = seq
        self.name = name
        self.constraints = constraints
        self.batch_size = batch_size
        self.w_ptm = w_ptm
        self.w_plddt = w_plddt
        self.w_globularity = w_globularity
        self.outdir = outdir
        self.verbose = verbose
        self.w_sasa = w_sasa
        self.w_bb_coord = w_bb_coord
        self.w_all_atm = w_all_atm

        # Parameters
        self.ref_pdbs = None

    def __str__(self):
        lines = [
            "ProteusAI.MCMC.ZeroShot class: \n",
            "---------------------------------------\n",
            "When Hallucination.run() sequences will be hallucinated using this seed sequence:\n\n",
            f"{self.seq}\n",
            "\nThe following variables were set:\n\n",
            "variable\t|value\n",
            "----------------+-------------------\n",
            f"batch_size: \t|{self.batch_size}\n",
            "The energy function is a linear combination of the following constraints:\n\n",
            "constraint\t|value\t|weight\n",
            "----------------+-------+------------\n",
        ]
        s = "".join(lines)
        lines = [
            s,
            f"pTM\t\t|\t|{self.w_ptm}\n",
            f"pLDDT\t\t|\t|{self.w_plddt}\n",
            f"bb_coord\t\t|{self.w_bb_coord}\n",
            f"all_atm\t\t|\t|{self.w_all_atm}\n",
            f"sasa\t\t|\t|{self.w_sasa}\n",
        ]
        s = "".join(lines)
        return s

    ### SAMPLERS
    def mutate(self, seq, pos):
        """
        mutates input sequences. to all possible amino acids at position

        Parameters:
            seqs (str): native protein sequence
            pos (int): position to mutate

        Returns:
            list: mutated sequences
        """

        AAs = (
            "A",
            "C",
            "D",
            "E",
            "F",
            "G",
            "H",
            "I",
            "K",
            "L",
            "M",
            "N",
            "P",
            "Q",
            "R",
            "S",
            "T",
            "V",
            "W",
            "Y",
        )

        native_aa = seq[pos]

        mut_seqs = []
        names = []
        for res in AAs:
            if res != native_aa:
                mut_seq = "".join([seq[:pos], res, seq[pos + 1 :]])
                mut_seqs.append(mut_seq)
                names.append(str(seq[pos]) + str(pos) + str(res))
        return mut_seqs, names

    ### ENERGY FUNCTION and ACCEPTANCE CRITERION
    def energy_function(self, seqs: list, pos: int, names):
        """
        Combines constraints into an energy function. The energy function
        returns the energy values of the mutated files and the associated pdb
        files as temporary files. In addition it returns a dictionary of the different
        energies.

        Parameters:
            seqs (list): list of sequences
            pos (int): position in sequence
            constraints (list): list of constraints

        Returns:
            tuple: Energy value, pdbs, energy_log
        """
        # reinitialize energy
        energies = np.zeros(len(seqs))
        energy_log = dict()
        constraints = self.constraints

        for i, c in enumerate(constraints["all_atm"]):
            if c == pos:
                del constraints["all_atm"][i]
        constraints = [constraints for i in range(len(seqs))]

        # structure prediction
        names = [f"{name}_{self.name}" for name in names]
        headers, sequences, pdbs, pTMs, mean_pLDDTs = Constraints.structure_prediction(
            seqs, names
        )
        pTMs = [1 - val for val in pTMs]
        mean_pLDDTs = [1 - val / 100 for val in mean_pLDDTs]

        e_pTMs = self.w_ptm * np.array(pTMs)
        e_mean_pLDDTs = self.w_plddt * np.array(mean_pLDDTs)
        e_globularity = self.w_globularity * Constraints.globularity(pdbs)
        e_sasa = self.w_sasa * Constraints.surface_exposed_hydrophobics(pdbs)

        energies += e_pTMs
        energies += e_mean_pLDDTs
        energies += e_globularity
        energies += e_sasa

        energy_log[f"e_pTMs x {self.w_ptm}"] = e_pTMs
        energy_log[f"e_mean_pLDDTs x {self.w_plddt}"] = e_mean_pLDDTs
        energy_log[f"e_globularity x {self.w_globularity}"] = e_globularity
        energy_log[f"e_sasa x {self.w_sasa}"] = e_sasa

        # there are now ref pdbs before the first calculation
        if self.ref_pdbs is not None:
            e_bb_coord = self.w_bb_coord * Constraints.backbone_coordination(
                pdbs, self.ref_pdbs
            )
            e_all_atm = self.w_all_atm * Constraints.all_atom_coordination(
                pdbs, self.ref_pdbs, constraints, constraints
            )

            energies += e_bb_coord
            energies += e_all_atm

            energy_log[f"e_bb_coord x {self.w_bb_coord}"] = e_bb_coord
            energy_log[f"e_all_atm x {self.w_all_atm}"] = e_all_atm
        else:
            energy_log[f"e_bb_coord x {self.w_bb_coord}"] = [0]
            energy_log[f"e_all_atm x {self.w_all_atm}"] = [0]

        energy_log["position"] = [pos]

        return energies, pdbs, energy_log

    ### RUN
    def run(self):
        """
        Runs MCMC-sampling based on user defined inputs. Returns optimized sequences.
        """
        seq = self.seq
        batch_size = self.batch_size
        energy_function = self.energy_function
        outdir = self.outdir
        mutate = self.mutate
        pdb_out = os.path.join(outdir, "pdbs")
        png_out = os.path.join(outdir, "pngs")
        data_out = os.path.join(outdir, "data_tools")

        if outdir is not None:
            if not os.path.exists(outdir):
                os.mkdir(outdir)
            if not os.path.exists(pdb_out):
                os.mkdir(pdb_out)
            if not os.path.exists(png_out):
                os.mkdir(png_out)
            if not os.path.exists(data_out):
                os.mkdir(data_out)

        if seq is None:
            raise "Provide a sequence(seq = <your_sequence>)"

        # for initial calculation don't use the full sequences, unecessary
        # calculation of initial state
        _, pdbs, energy_log = energy_function([seq], 0, ["native"])
        pdbs = [pdbs[0] for _ in range(batch_size - 1)]
        energy_log["mut"] = ["-"]
        energy_log["description"] = ["native"]

        # make energies to list
        for key in energy_log.keys():
            if not isinstance(energy_log[key], list):
                energy_log[key] = energy_log[key].tolist()

        self.ref_pdbs = pdbs.copy()

        if outdir is not None:
            # saves the n th structure
            pdbs[0].write(os.path.join(pdb_out, f"native_{self.name}.pdb"))

        # write energy_log in data_out
        if outdir is not None:
            df = pd.DataFrame(energy_log)
            df.to_csv(os.path.join(data_out, "energy_log.pdb"), index=False)

        for pos in range(len(seq)):
            seqs, names = mutate(seq, pos)
            E_x_i, pdbs, _energy_log = energy_function(seqs, pos, names)

            for n in range(len(seqs)):
                for key in energy_log.keys():
                    # skip skalar values in this step
                    if key not in ["position", "mut", "description"]:
                        e = _energy_log[key].tolist()
                        with open("test", "w") as f:
                            print("energy_log", file=f)
                            print(energy_log, file=f)
                            print("_energy_log", file=f)
                            print(_energy_log, file=f)
                            print("key", file=f)
                            print(key, file=f)
                            print("e", file=f)
                            print(e, file=f)
                        energy_log[key].append(e[n])

                energy_log["position"].append(pos)
                energy_log["mut"].append(names[n])

                energy_log["description"].append(f"{names[n]}_{self.name}")

                if outdir is not None:
                    # saves the n th structure
                    pdbs[n].write(os.path.join(pdb_out, f"{names[n]}_{self.name}.pdb"))

                # write energy_log in data_out
                if outdir is not None:
                    df = pd.DataFrame(energy_log)
                    df.to_csv(os.path.join(data_out, "energy_log.pdb"), index=False)



import numpy as np
import random


#####################################
### Simulated Annealing Discovery ###
#####################################


def precompute_distances(vectors):
    """Precompute the pairwise Euclidean distance matrix."""
    num_vectors = len(vectors)
    distance_matrix = np.zeros((num_vectors, num_vectors))

    for i in range(num_vectors):
        for j in range(i + 1, num_vectors):
            dist = np.linalg.norm(vectors[i] - vectors[j])
            distance_matrix[i, j] = dist
            distance_matrix[j, i] = dist

    return distance_matrix


def diversity_score_incremental(
    current_score, selected_indices, idx_in, idx_out, distance_matrix
):
    """Update the diversity score incrementally when a vector is swapped."""
    new_score = current_score

    for idx in selected_indices:
        if idx != idx_out:
            new_score -= distance_matrix[idx_out, idx]
            new_score += distance_matrix[idx_in, idx]

    return new_score


def simulated_annealing(
    vectors,
    N,
    initial_temperature=1000.0,
    cooling_rate=0.003,
    max_iterations=10000,
    pbar=None,
):
    """
    Simulated Annealing to select N vectors that maximize diversity.

    Args:
        vectors (list): List of numpy arrays.
        N (int): Number of sequences that should be sampled.
        initial_temperature (float): Initial temperature of the simulated annealing algorithm. Default 1000.0.
        cooling_rate (float): Cooling rate of the simulated annealing algorithm. Default 0.003.
        max_iterations (int): Maximum number of iterations of the simulated annealing algorithm. Default 10000.

    Returns:
        list: Indices of diverse vectors.
    """

    if pbar:
        pbar.set(message="Computing distance matrix", detail="...")

    # Precompute all pairwise distances
    distance_matrix = precompute_distances(vectors)

    # Randomly initialize the selection of N vectors
    selected_indices = random.sample(range(len(vectors)), N)
    current_score = sum(
        distance_matrix[i, j]
        for i in selected_indices
        for j in selected_indices
        if i < j
    )

    temperature = initial_temperature
    best_score = current_score
    best_selection = selected_indices[:]

    for iteration in range(max_iterations):

        if pbar:
            pbar.set(iteration, message="Minimizing energy", detail="...")

        # Randomly select a vector to swap
        idx_out = random.choice(selected_indices)
        idx_in = random.choice(
            [i for i in range(len(vectors)) if i not in selected_indices]
        )

        # Incrementally update the diversity score
        new_score = diversity_score_incremental(
            current_score, selected_indices, idx_in, idx_out, distance_matrix
        )

        # Decide whether to accept the new solution
        delta = new_score - current_score
        if delta > 0 or np.exp(delta / temperature) > random.random():
            selected_indices.remove(idx_out)
            selected_indices.append(idx_in)
            current_score = new_score

            # Update the best solution found so far
            if new_score > best_score:
                best_score = new_score
                best_selection = selected_indices[:]

        # Cool down the temperature
        temperature *= 1 - cooling_rate

        # Early stopping if the temperature is low enough
        # if temperature < 1e-8:
        #    break

    return best_selection, best_score


#######################################
### Genetic Algorithm for Mutations ###
#######################################


def find_mutations(sequences):
    """
    Takes a list of protein sequences and returns a dictionary with mutation positions
    as keys and lists of amino acids at those positions as values.

    Parameters:
    sequences (list): A list of protein sequences (strings).

    Returns:
    dict: A dictionary where keys are positions (1-indexed) and values are lists of amino acids at those positions.
    """
    # Check if the list is empty
    if not sequences:
        return {}

    # Initialize a dictionary to store mutations
    mutations = {}

    # Get the reference sequence (assuming all sequences have the same length)
    reference_seq = sequences[0]

    # Iterate through each position in the sequence
    for i in range(len(reference_seq)):
        # Initialize a set to store the different amino acids at this position
        amino_acids = set()

        # Check the amino acid at this position in each sequence
        for seq in sequences:
            amino_acids.add(seq[i])

        # If there is more than one unique amino acid at this position, it's a mutation
        if len(amino_acids) > 1:
            mutations[i + 1] = list(amino_acids)  # +1 to make position 1-indexed

    return mutations



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

import os
import sys
import proteusAI.ml_tools.esm_tools.esm_tools as esm_tools
import proteusAI.ml_tools.torch_tools as torch_tools
import proteusAI.io_tools as io_tools
import proteusAI.visual_tools as vis
import proteusAI.struc as pai_struc
import pandas as pd
from proteusAI.Protein.protein import Protein
from typing import Union, Optional
from sklearn.preprocessing import LabelEncoder

current_path = os.path.dirname(os.path.abspath(__file__))
root_path = os.path.join(current_path, "..")
sys.path.append(root_path)
folder_path = os.path.dirname(os.path.realpath(__file__))
USR_PATH = os.path.join(folder_path, "../../../usrs")


class Library:
    """
    The Library object holds information about proteins, labels and representations.
    It is also used to create mathematical representations of proteins.

    The library object serves as input to Model objects, to train machine learning models.

    Attributes:
        usr (str): User name - determines where data will be stored. default: Guest
        data (str): Path to data file ().
        proteins (list): List of proteins.
    """

    # TODO: add VAEs too

    representation_types = ["esm1v", "esm2", "ohe", "blosum62", "blosum50", "vae"]
    _allowed_y_types = ["class", "num"]
    in_memory = ["ohe", "blosum62", "blosum50"]

    def __init__(
        self,
        user: str = "guest",
        source: Union[str, dict, None] = None,
        seqs_col: Union[str, None] = None,
        names_col: Union[str, None] = None,
        y_col: Union[str, None] = None,
        y_type: Union[str, None] = None,
        sheet: Union[str, None] = None,
        fname: Union[str, None] = None,
    ):
        """
        Initialize a new library.

        Args:
            user (str): User name.
            source (str, or dict): Source of data, either a file or a data package created from a diversification step.
            seqs_col (str): Column name for the sequences, if the source is a '.csv' or Excel file.
            names_col (str): Column name for the names (sequence descriptions), if the source is a '.csv' or Excel file.
            y_col (str): Column name for the y-values (e.g. fitness, stability, activity), if the source is a '.csv' or Excel file.
            y_type (str): Specify the data type of y-values, either categorical or numerical. Will parse automatically if none is provided.
            sheet (str): Specify the excel sheet name, if the source is an Excel file.
            fname (str): Only relevant for the app - provides the real file name instead of temporary file name from shiny.

        Parameters:
            data (df): dataframe of uploaded data (raw data).
            seqs (list): list of sequence.
            y (list): list of y values.
            y_pred (list): list of predicted y values.
            y_type (str): Type of y values ('class' or 'num').
            names (list): names of sequences.
            reps (list): list of computed representations.
            proteins (list): list of protein objects.
            pred_data (bool): Using predicted data, e.g. predicted ZS y-values with no real y-values
        """
        # Arguments
        self.user = os.path.join(USR_PATH, user)
        self.source = source
        self.seq_col = seqs_col
        self.names_col = names_col
        self.y_col = y_col
        self.y_type = y_type
        self.sheet = sheet
        self.fname = fname

        # Parameters
        self.data = None
        self.seqs = None
        self.y = None
        self.y_pred = None
        self.names = None
        self.reps = self.in_memory.copy()
        self.source_path = None
        self.rep_path = None
        self.strucs = None
        self.struc_path = None
        self.class_dict = None
        self.pred_data = False

        # Create user if user does not exist
        if not os.path.exists(self.user):
            self.initialize_user()

        # Initialize library from file or from inheritance
        if isinstance(self.source, str):
            self.init_from_file()
        else:
            self.init_from_inheritance()

    def init_from_file(self):
        """
        Initialize a library from a file.
        """

        # handle app case
        if self.fname:
            fname = self.fname.split(".")[0]
            file_extension = self.fname.split(".")[-1]
        else:
            f = self.source.split("/")[-1]
            fname = f.split(".")[0]
            file_extension = f.split(".")[-1]

        # set paths
        self.source_path = os.path.join(self.user, fname)
        self.rep_path = os.path.join(self.source_path, "library/rep")
        self.struc_path = os.path.join(self.source_path, "library/struc")

        # create user library if user does not exist
        os.makedirs(self.source_path, exist_ok=True)
        os.makedirs(self.source_path, exist_ok=True)
        os.makedirs(os.path.join(self.source_path, "library/rep"), exist_ok=True)
        os.makedirs(os.path.join(self.source_path, "library/struc"), exist_ok=True)

        # load the data
        if file_extension in ["xlsx", "xls", "csv"]:
            self._read_tabular_data(
                in_file=self.source,
                seqs=self.seq_col,
                y=self.y_col,
                y_type=self.y_type,
                names=self.names_col,
                sheet=self.sheet,
                file_ext=file_extension,
            )
        elif file_extension in ["fasta", "fa"]:
            self._read_fasta(self.source)
        else:
            raise ValueError(f"Unsupported file type: {file_extension}")

    def init_from_inheritance(self):
        """
        Initialize a library from another process.

        Args:
            data (dict): Contains parent path and path to data file.
        """

        # For the future adapt this so it can handle multi-factor optimization
        # extract data from parent file
        data = self.source

        # Parsing parameters
        self.rep_path = data["rep_path"]
        self.struc_path = data["struc_path"]
        self.seq_col = data["seqs_col"]
        self.names_col = data["names_col"]
        self.reps = data["reps"]
        self.class_dict = data["class_dict"]

        # Parsing arguments
        df = data["df"]
        self.seqs = df[self.seq_col].to_list()
        self.names = df[self.names_col].to_list()
        self.y_type = data["y_type"]
        self.data = df

        # if true y values are there
        if "y_col" in data.keys():
            self.y_col = data["y_col"]
            self.y = df[self.y_col].to_list()
        else:
            self.y = [None] * len(self.seqs)

        # if predicted y values are there
        if "y_pred_col" in data.keys():
            self.y_pred_col = data["y_pred_col"]
            self.y_pred = df[self.y_pred_col].to_list()
        else:
            self.y_pred = [None] * len(self.seqs)

        # if predicted y values are there
        if "y_sigma_col" in data.keys():
            self.y_sigma_col = data["y_sigma_col"]
            self.y_sigma = df[self.y_sigma_col].to_list()
        else:
            self.y_sigma = [None] * len(self.y)

        # if predicted y values are there
        if "acq_col" in data.keys():
            self.acq_col = data["acq_col"]
            self.acq_score = df[self.acq_col].to_list()
        else:
            self.acq_score = [None] * len(self.y)

        if "pred_data" in data.keys():
            self.pred_data = data["pred_data"]

        # create proteins
        self.proteins = [
            Protein(
                name,
                seq,
                y=y,
                y_pred=y_pred,
                y_sigma=y_sigma,
                acq_score=acq_score,
                user=self.user,
            )
            for name, seq, y, y_pred, y_sigma, acq_score in zip(
                self.names, self.seqs, self.y, self.y_pred, self.y_sigma, self.acq_score
            )
        ]

    def initialize_user(self):
        """
        initializing a new library.
        """
        print(f"Initializing user '{self.user}'...")

        # create user library if user does not exist
        if not os.path.exists(self.user):
            os.makedirs(self.user)
            if self.fname and not os.path.exists(
                os.path.join(self.user, self.fname.split(".")[0])
            ):
                fname = self.fname.split(".")[0]
                os.makedirs(os.path.join(self.user, f"{fname}"))
                os.makedirs(os.path.join(self.user, f"{fname}/library"))
                os.makedirs(os.path.join(self.user, f"{fname}/zero_shot"))
                os.makedirs(os.path.join(self.user, f"{fname}/design"))
            print(f"User created at {self.user}")

        # check if sequence have been provided
        if len(self.seqs) > 0:
            # create dummy names if no names are provided
            if len(self.seqs) != len(self.names):
                print(
                    f"Number of sequences ({len(self.seqs)}), does not match the number of names ({len(self.names)})"
                )
                print("Dummy names will be created")
                self.names = [f"protein_{i}" for i in range(len(self.seqs))]

            # create protein objects TODO: This is the slow step
            if len(self.y) == len(self.names):
                for name, seq, y in zip(self.names, self.seqs, self.y):
                    protein = Protein(name, seq, y=y)
                    self.proteins.append(protein)

            else:
                for name, seq in zip(self.names, self.seqs):
                    protein = Protein(name, seq)
                    self.proteins.append(protein)

        print("Done!")

    ### IO ###
    def load_library(self):
        """
        Checking the availability of precomputed representations.
        """
        print(f"Loading library '{self.user}'...")

        # Check for representations
        if self.rep_path is None:
            if self.file:
                fname = self.file.split(".")[0]
                rep_path = os.path.join(self.user, f"{fname}/library/rep")
                struc_path = os.path.join(self.user, f"{fname}/library/struc")
                self.rep_path = rep_path
                self.struc_path = struc_path
            else:
                rep_path = os.path.join(self.user, "rep")
                struc_path = os.path.join(self.user, "struc")
                self.rep_path = rep_path
                self.struc_path = struc_path
        else:
            rep_path = "/".join(self.rep_path.split("/")[:-1])

        if os.path.exists(rep_path):
            for rep_type in self.representation_types:
                rep_type_path = os.path.join(rep_path, rep_type)
                if os.path.exists(rep_type_path):
                    self.reps.append(rep_type)
                    print(
                        f"- Found representations of type '{rep_type}' in 'rep/{rep_type}'."
                    )

        print("Loading done!")

    def read_data(
        self,
        data: str,
        seqs: Union[str, None] = None,
        y: Union[str, None] = None,
        y_type: str = "num",
        names: Union[str, None] = None,
        sheet: Optional[str] = None,
    ):
        """
        Reads data from a CSV, Excel, or FASTA file and populates the Library object.

        Args:
            data (str): Path to the data file. Tabular data or fasta
            seqs (str): Column name for sequences in the data file.
            y (str): Column name for y values in the data file.
            y_type (str): Type of y values ('class' or 'num').
            names (str, optional): Column name for sequence names in the data file.
            sheet (str, optional): Name of the Excel sheet to read.
        """

        assert y_type in self._allowed_y_types
        self.y_type = y_type

        # Determine file type based on extension
        file_ext = os.path.splitext(data)[1].lower()
        self.file = data.split("/")[-1].split(".")[0]

        self.rep_path = os.path.join(self.user, self.file, "library/rep")
        self.struc_path = os.path.join(self.user, self.file, "library/struc")

        if file_ext in [".xlsx", ".xls", ".csv"]:
            self._read_tabular_data(
                in_file=data,
                seqs=seqs,
                y=y,
                y_type=y_type,
                names=names,
                sheet=sheet,
                file_ext=file_ext,
            )
        elif file_ext in [".fasta", ".fa"]:
            self._read_fasta(data)
        else:
            raise ValueError(f"Unsupported file type: {file_ext}")

    def _read_tabular_data(
        self,
        in_file: str,
        seqs: str,
        y: Union[str, None],
        y_type: Union[str, None],
        names: Union[str, None],
        sheet: Optional[str],
        file_ext: Union[str, None] = None,
        check_rep: bool = True,
        check_strucs: bool = True,
    ):
        """
        Reads data from a CSV, Excel, and populates the Library object. Called by read_data

            Args:
                data (str): Path to the data file.
                seqs (str): Column name for sequences in the data file.
                y (str): Column name for y values in the data file.
                y_type (str): Type of y values ('class' or 'num').
                names (str, optional): Column name for sequence names in the data file.
                sheet (str, optional): Name of the Excel sheet to read.
                file_ext (str): file extension
                check_rep (bool): Check if all representations have been computed.
                check_sturcs (bool): Check if structures have been computed.
        """
        if file_ext in ["xlsx", "xls"]:
            if sheet is None:
                df = pd.read_excel(in_file)
            else:
                df = pd.read_excel(in_file, sheet_name=sheet)
        elif file_ext == "csv":
            df = pd.read_csv(in_file)
        else:
            raise ValueError(f"Unsupported file type: {file_ext}")

        self.data = df
        # Validate the columns exist
        if seqs not in df.columns or y not in df.columns:
            raise ValueError(
                "The provided column names do not match the columns in the data file."
            )

        # If names are not provided, generate dummy names
        if names not in df.columns:
            self.names = [f"protein_{i}" for i in range(len(self.seqs))]
        else:
            self.names = df[names].tolist()

        self.seqs = df[seqs].tolist()

        # Handle y values
        if y is not None:
            self.y = df[y].tolist()

            if y_type == "class":
                self.y, self.class_dict = self._encode_categorical_labels(self.y)

            # Create protein objects with y values
            self.proteins = [
                Protein(name, seq, y=y, user=self.user)
                for name, seq, y in zip(self.names, self.seqs, self.y)
            ]
        else:
            # Create protein objects without y values
            self.proteins = [
                Protein(name, seq, user=self.user)
                for name, seq in zip(self.names, self.seqs)
            ]

        if check_rep:
            self._check_reps()

        if check_strucs:
            self._check_strucs()

    def _read_fasta(self, in_file, check_rep: bool = False):
        """
        Read fasta file and create protein objects.

        Args:
            data (str): Path to data.
            check_rep (bool): Check if all representations have been computed.
        """
        names, sequences = io_tools.fasta.load_fasta(in_file)
        self.names = names
        self.seqs = sequences

        # Create protein objects from names and sequences
        self.proteins = [Protein(name, seq) for name, seq in zip(self.names, self.seqs)]
        self.y = [None] * len(self.proteins)
        df = pd.DataFrame({"names": names, "sequence": self.seqs, "y": self.y})
        self.data = df

        if check_rep:
            self._check_reps()

    def _check_reps(self):
        """
        Check for available representations, store in protein object if representation is found
        """
        reps = [r for r in os.listdir(self.rep_path) if not r.startswith(".")]
        if len(reps) > 0:
            for rep in reps:
                rep_names = [
                    f
                    for f in os.listdir(os.path.join(self.rep_path, rep))
                    if f.endswith(".pt")
                ]
                if len(rep_names) >= len(set(self.names)):
                    self.reps.insert(0, rep)  # Insert representation at the front
                    for protein in self.proteins:
                        f_name = protein.name + ".pt"
                        if os.path.exists(os.path.join(self.rep_path, rep, f_name)):
                            protein._reps.insert(0, rep)

    def _check_strucs(self):
        """
        Check for available representations, store in protein object if representation is found
        """
        strucs = [r for r in os.listdir(self.struc_path) if not r.startswith(".")]
        if len(strucs) > 0:
            for struc in strucs:
                struc_names = [
                    f
                    for f in os.path.join(self.struc_path, struc)
                    if f.endswith(".pdb")
                ]
                if len(struc_names) == len(set(self.names)):
                    self.strucs.append(struc)
                    for protein in self.proteins:
                        f_name = protein.name + ".pdb"  # noqa: F841
                        protein._strucss.append(strucs)

    def _encode_categorical_labels(self, ys):
        """
        Encode categorical labels into numerical format.
        """
        label_encoder = LabelEncoder()
        encoded_labels = label_encoder.fit_transform(ys).tolist()
        class_mapping = {
            index: label for index, label in enumerate(label_encoder.classes_)
        }
        return encoded_labels, class_mapping

    ### Utility ###
    def rename_proteins(self, new_names: Union[list, tuple]):
        """
        Renames all proteins in the Library object with provided names.

        Args:
            new_names (list or tuple): List of new protein names.
        """

        if len(new_names) != len(self.proteins):
            raise ValueError(
                f"Number of provided names ({len(new_names)}) does not match number of proteins in the library ({len(self.proteins)})."
            )

        # Update names of Protein objects and in the self.names list
        for protein, new_name in zip(self.proteins, new_names):
            protein.name = new_name

        self.names = new_names

        self._check_reps()
        self._check_strucs()

    def set_y_values(self, y_values: list):
        """
        Sets the y values for all proteins.

        Args:
            y_values (list): List of y values.
        """

        if len(y_values) != len(self.proteins):
            raise ValueError(
                f"Number of provided y values ({len(y_values)}) does not match number of proteins in the library ({len(self.proteins)})."
            )

        for protein, y_value in zip(self.proteins, y_values):
            protein.y = y_value

    ### Representation builders ###
    def compute(
        self,
        method: str,
        batch_size: int = 100,
        dest: Union[str, None] = None,
        pbar=None,
        device=None,
        proteins=None,
    ):
        """
        Compute representations for proteins.

        Args:
            method (str): Method for computing representation
            batch_size (int, optional): Batch size for representation computation.
            dest (str): destination of representations
            pbar: Progress bar for shiny app.
            device (str): Choose hardware for computation. Default 'None' for autoselection
                          other options are 'cpu' and 'cuda'.
            proteins (list): list of specific proteins. Optional
        """
        simple_rep_types = ["ohe", "blosum62", "blosum50"]
        supported_methods = self.representation_types + simple_rep_types

        assert method in supported_methods, f"'{method}' is not a supported method"
        assert isinstance(batch_size, (int, type(None)))

        if method in ["esm2", "esm1v"]:
            self.esm_builder(
                model=method, batch_size=batch_size, dest=dest, pbar=pbar, device=device
            )
        elif method == "ohe":
            reps = self.ohe_builder(dest=dest, pbar=pbar, proteins=proteins)
            return reps
        elif method in ["blosum62", "blosum50"]:
            reps = self.blosum_builder(
                matrix_type=method.upper(), dest=dest, pbar=pbar, proteins=proteins
            )
            return reps

    def esm_builder(
        self,
        model: str = "esm2",
        batch_size: int = 10,
        dest: Union[str, None] = None,
        pbar=None,
        device=None,
    ):
        """
        Computes esm representations.

        Args:
            model (str): Supports esm2 and esm1v.
            batch_size (int): Batch size for computation.
            dest (str): destination of representations
            pbar: Progress bar for shiny app.
            device (str): Choose hardware for computation. Default 'None' for autoselection
                          other options are 'cpu' and 'cuda'.
        """

        dest = os.path.join(self.rep_path, model)

        if not os.path.exists(dest):
            os.makedirs(dest)

        # Filtering out proteins that have already computed representations
        proteins_to_compute = [
            protein
            for protein in self.proteins
            if not os.path.exists(os.path.join(dest, protein.name + ".pt"))
        ]

        print(f"computing {len(proteins_to_compute)} proteins")

        if pbar:
            pbar.set(
                message=f"Computing {len(proteins_to_compute)} representations",
                detail="...",
            )

        # get names for and sequences for computation
        names = [protein.name for protein in proteins_to_compute]
        seqs = [protein.seq for protein in proteins_to_compute]

        # compute representations
        esm_tools.batch_compute(
            seqs,
            names,
            dest=dest,
            model=model,
            batch_size=batch_size,
            pbar=pbar,
            device=device,
        )

        for protein in proteins_to_compute:
            if model not in protein.reps:
                protein.reps.append(model)

        if model not in self.reps:
            self.reps.append(model)

    def ohe_builder(self, dest: Union[str, None] = None, pbar=None, proteins=None):
        """
        Computes one-hot encoding representations for proteins using one_hot_encoder method.
        Assumes all data fits in memory.

        Args:
            dest (str): destination of representations
        """
        if proteins:
            seqs = [prot.seq for prot in proteins]
        else:
            seqs = self.seqs

        # Determine the maximum sequence length for padding
        max_sequence_length = max(len(seq) for seq in self.seqs)

        # Compute the one-hot encoding with the calculated padding
        ohe_representations = torch_tools.one_hot_encoder(
            seqs, pbar=pbar, padding=max_sequence_length
        )

        return ohe_representations

    def blosum_builder(
        self,
        matrix_type="BLOSUM62",
        dest: Union[str, None] = None,
        pbar=None,
        proteins=None,
    ):
        """
        Computes BLOSUM representations for proteins using blosum_encoding method.
        Assumes all data fits in memory.

        Args:
            matrix_type (str): Type of BLOSUM matrix to use.
            dest (str): destination
        """
        if proteins:
            seqs = [prot.seq for prot in proteins]
        else:
            seqs = self.seqs

        # Determine the maximum sequence length for padding
        max_sequence_length = max(len(seq) for seq in self.seqs)

        # Compute the BLOSUM encoding with the calculated padding
        blosum_representations = torch_tools.blosum_encoding(
            seqs, matrix=matrix_type, pbar=pbar, padding=max_sequence_length
        )

        return blosum_representations

    def load_representations(
        self, rep: Union[str, None], proteins: Union[list, None] = None
    ):
        """
        Loads representations for a list of proteins.

        Args:
            rep (str): type of representation to load
            proteins (list): list of proteins to load, load all if None

        Returns:
            list: List of representations.
        """

        if self.rep_path is None:
            rep_path = os.path.join(self.user, f"rep/{rep}")
        else:
            rep_path = os.path.join(self.rep_path, rep)

        if proteins is None:
            file_names = [protein.name + ".pt" for protein in self.proteins]
        else:
            file_names = [protein.name + ".pt" for protein in proteins]

        if rep in self.in_memory:
            reps = self.compute(method=rep, proteins=proteins)
        else:
            _, reps = io_tools.load_embeddings(path=rep_path, names=file_names)
        return reps

    ### Folding ###
    def fold(
        self,
        names,
        model: str = "esm_fold",
        num_recycles: int = 0,
        pbar=None,
        relax: bool = False,
    ):
        """
        Fold sequences by their names.

        Args:
            names (list): list of names to fold
            model (str): model used for folding, currently only esm_fold
            num_recycles (int): number of recycling steps, default 0
            pbar: progress bar for app
        """
        out = False

        seqs = self.data[self.data[self.names_col].isin(names)][self.seq_col].to_list()
        if len(seqs) > 0:
            all_headers, all_sequences, all_pdbs, pTMs, mean_pLDDTs = (
                esm_tools.structure_prediction(
                    names=names, seqs=seqs, num_recycles=num_recycles, pbar=pbar
                )
            )

            if not os.path.exists(self.struc_path):
                os.makedirs(self.struc_path)

            for i, name in enumerate(names):
                dest = os.path.join(self.struc_path, name + ".pdb")
                f = all_pdbs[i]
                f.write(dest)

            if relax:
                if pbar:
                    pbar.set(
                        message="Initiating energy minimization",
                        detail=f"Minimizing {len(names)} structures...",
                    )
                for i, name in enumerate(names):
                    if pbar:
                        pbar.set(
                            i + 1,
                            message="Minimizing energy",
                            detail=f"{i+1}/{len(names)} remaining...",
                        )
                    self.relax_struc(name)

            df = pd.DataFrame(
                {
                    "name": all_headers,
                    "sequence": all_sequences,
                    "pLDDT": mean_pLDDTs,
                    "pTM": pTMs,
                }
            )
            out = {
                "df": df,
                "rep_path": self.rep_path,
                "struc_path": self.struc_path,
                "y_type": "num",
                "y_col": "pLDDT",
                "seqs_col": "sequence",
                "names_col": "name",
                "reps": self.reps,
                "class_dict": self.class_dict,
            }

        return out

    ### Structure ###
    def relax_struc(self, name: str):
        """
        Perform energy minimization on a protein structure.
        """
        f = os.path.join(self.struc_path, name + ".pdb")
        pai_struc.relax_pdb(f, dest=self.struc_path)

    def struc_geom(self, ref, residues: dict = {}):
        """
        Compute the difference in geometry of a dictionary of residues in a library compared to a reference structure.
        The residues must be present in both structures, or they won't be compared.

        Args:
            ref (proteusAI.Protein): Reference protein structure.
            residues (dict): Dictionary where keys are chain identifiers and values are lists of residues to compare.

        Returns:
            A DataFrame containing comparison results for each protein in the library.
        """

        ref_struc = ref.struc
        ref_chi = pai_struc.compute_chi_angles(ref_struc, residues)

        rmsds = []
        delta_chis = []

        for prot in self.proteins:
            target_path = os.path.join(self.struc_path, prot.name + ".pdb")
            ref_struc, target = pai_struc.align(ref.pdb_file, target_path)
            rmsd = pai_struc.compute_rmsd(ref_struc, target)
            target_chi = pai_struc.compute_chi_angles(target, residues)

            # Calculate delta chi for each chain and residue specified
            delta_chi = {}
            for chain, res_list in residues.items():
                for res_id in res_list:
                    if (chain, res_id) in ref_chi and (chain, res_id) in target_chi:
                        # Compute difference for each chi angle
                        chi_differences = [
                            abs(ref_angle - tar_angle)
                            for ref_angle, tar_angle in zip(
                                ref_chi[(chain, res_id)], target_chi[(chain, res_id)]
                            )
                        ]
                        delta_chi[(chain, res_id)] = chi_differences

            delta_chis.append(delta_chi)
            rmsds.append(rmsd)

        results_path = os.path.join(self.rep_path, f"{prot.name}_analysis.csv")

        if not os.path.exists(self.rep_path):
            os.makedirs(self.rep_path, exist_ok=True)

        df = pd.DataFrame(
            {
                "names": [prot.name for prot in self.proteins],
                "RMSD": rmsds,
                "delta_chi": delta_chis,
            }
        )
        df.to_csv(results_path)

        return df

    def plot_tsne(
        self,
        rep: str,
        y_upper=None,
        y_lower=None,
        names=None,
        highlight_mask=None,
        highlight_label=None,
    ):
        """
        Plot representations with optional thresholds and point names.

        Args:
            rep (str): Representation type to plot.
            y_upper (float, optional): Upper threshold for special coloring.
            y_lower (float, optional): Lower threshold for special coloring.
            names (List[str], optional): List of names for each point.
            highlight_mask (list): List of 0s and 1s to highlight plot. Default None.
            highlight_label (str): Text for the legend entry of highlighted points.
        """

        x = self.load_representations(rep)
        y = self.y

        if self.y_type == "class":
            y = [self.class_dict[i] for i in y]

        fig, ax, df = vis.plot_tsne(
            x,
            y,
            y_upper=y_upper,
            y_lower=y_lower,
            names=names,
            rep_type=rep,
            y_type=self.y_type,
            random_state=42,
            highlight_mask=highlight_mask,
            highlight_label=highlight_label,
        )

        return fig, ax, df

    def plot_umap(
        self,
        rep: str,
        y_upper=None,
        y_lower=None,
        names=None,
        highlight_mask=None,
        highlight_label=None,
    ):
        """
        Plot representations with optional thresholds and point names.

        Args:
            rep (str): Representation type to plot.
            y_upper (float, optional): Upper threshold for special coloring.
            y_lower (float, optional): Lower threshold for special coloring.
            names (List[str], optional): List of names for each point.
            highlight_mask (list): List of 0s and 1s to highlight plot. Default None.
            highlight_label (str): Text for the legend entry of highlighted points.
        """

        x = self.load_representations(rep)
        y = self.y

        if self.y_type == "class":
            y = [self.class_dict[i] for i in y]

        fig, ax, df = vis.plot_umap(
            x,
            y,
            y_upper=y_upper,
            y_lower=y_lower,
            names=names,
            rep_type=rep,
            y_type=self.y_type,
            random_state=42,
            highlight_mask=highlight_mask,
            highlight_label=highlight_label,
        )

        return fig, ax, df

    def plot_pca(
        self,
        rep: str,
        y_upper=None,
        y_lower=None,
        names=None,
        highlight_mask=None,
        highlight_label=None,
    ):
        """
        Plot representations with optional thresholds and point names.

        Args:
            rep (str): Representation type to plot.
            y_upper (float, optional): Upper threshold for special coloring.
            y_lower (float, optional): Lower threshold for special coloring.
            names (List[str], optional): List of names for each point.
            highlight_mask (list): List of 0s and 1s to highlight plot. Default None.
            highlight_label (str): Text for the legend entry of highlighted points.
        """

        x = self.load_representations(rep)
        y = self.y

        if self.y_type == "class":
            y = [self.class_dict[i] for i in y]

        fig, ax, df = vis.plot_pca(
            x,
            y,
            y_upper=y_upper,
            y_lower=y_lower,
            names=names,
            rep_type=rep,
            y_type=self.y_type,
            random_state=42,
            highlight_mask=highlight_mask,
            highlight_label=highlight_label,
        )

        return fig, ax, df

    def __len__(self):
        return len(self.seqs)

    def top_n(self, n: int = 10, ascending=False):
        """
        Returns the top n-proteins by y-value.

        Args:
            n (int): Number of top variants. Default 10
            ascending (bool): sort in ascending or descending order
        """

        # Get the list of proteins
        proteins = self.proteins

        # Sort the proteins by y-value
        sorted_proteins = sorted(
            proteins, key=lambda protein: protein.y, reverse=not ascending
        )

        # Return the top n proteins
        return sorted_proteins[:n]



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk and Laura Sofia Machado"

import numpy as np
from scipy.stats import norm


def greedy(mean, std=None, current_best=None, xi=None):
    """
    Greedy acquisition function.

    Args:
        mean (np.array): This is the mean function from the GP over the considered set of points.
        std (np.array, optional): This is the standard deviation function from the GP over the considered set of points. Default is None.
        current_best (float, optional): This is the current maximum of the unknown function: mu^+. Default is None.
        xi (float, optional): Small value added to avoid corner cases. Default is None.

    Returns:
        np.array: The mean values for all the points, as greedy acquisition selects the best based on mean.
    """
    return mean


def EI(mean, std, current_best, xi=0.1):
    """
    Expected Improvement acquisition function.

    It implements the following function:

            | (mu - mu^+ - xi) Phi(Z) + sigma phi(Z) if sigma > 0
    EI(x) = |
            | 0                                       if sigma = 0

            where Phi is the CDF and phi the PDF of the normal distribution
            and
            Z = (mu - mu^+ - xi) / sigma

    Args:
        mean (np.array): This is the mean function from the GP over the considered set of points.
        std (np.array): This is the standard deviation function from the GP over the considered set of points.
        current_best (float): This is the current maximum of the unknown function: mu^+.
        xi (float): Small value added to avoid corner cases.

    Returns:
        np.array: The value of this acquisition function for all the points.
    """

    Z = (mean - current_best - xi) / (std + 1e-9)
    EI = (mean - current_best - xi) * norm.cdf(Z) + std * norm.pdf(Z)
    EI[std == 0] = 0

    return EI


def UCB(mean, std, current_best=None, kappa=1.5):
    """
    Upper-Confidence Bound acquisition function.

    Args:
        mean (np.array): This is the mean function from the GP over the considered set of points.
        std (np.array): This is the standard deviation function from the GP over the considered set of points.
        current_best (float, optional): This is the current maximum of the unknown function: mu^+. Default is None.
        kappa (float): Exploration-exploitation trade-off parameter. The higher the value, the more exploration. Default is 0.

    Returns:
        np.array: The value of this acquisition function for all the points.
    """
    return mean + kappa * std


def random_acquisition(mean, std=None, current_best=None, xi=None):
    """
    Random acquisition function. Assigns random acquisition values to all points in the unobserved set.

    Args:
        mean (np.array): This is the mean function from the GP over the considered set of points.
        std (np.array, optional): This is the standard deviation function from the GP over the considered set of points. Default is None.
        current_best (float, optional): This is the current maximum of the unknown function: mu^+. Default is None.
        xi (float, optional): Small value added to avoid corner cases. Default is None.

    Returns:
        np.array: Random acquisition values for all points in the unobserved set.
    """
    n_unobserved = len(mean)
    np.random.seed(None)
    random_acq_values = np.random.random(n_unobserved)
    return random_acq_values



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

import os
import torch
from typing import Union


def load_embeddings(
    path: str, names: Union[list, None] = None, map_location: str = "cpu"
) -> tuple:
    """
    Loads all representations files from a directory, returns the names/ids and sequences as lists.

    Parameters:
        path (str): path to directory containing representations files
        names (list): list of file names in case files should be loaded in a specific order if names not None.
            Will go to provided path and load files by name order. Default None

    Returns:
        tuple: two lists containing the names and sequences as torch tensors

    Example:
        names, sequences = load('/path/to/representations')
    """

    tensors = []
    if names is None:
        files = [os.path.join(path, f) for f in os.listdir(path) if f.endswith(".pt")]
        names = [f[:-3] for f in os.listdir(path) if f.endswith(".pt")]
        for f in files:
            t = torch.load(f, map_location=map_location)
            tensors.append(t)
    else:
        for name in names:
            t = torch.load(os.path.join(path, name), map_location=map_location)
            tensors.append(t)

    return names, tensors



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

"""
A subpackage for data_tools engineering, visualization and analysis.
"""

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from proteusAI.data_tools.pdb import *  # noqa: F403
from proteusAI.data_tools.MSA import *  # noqa: F403



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

"""
A subpackage for mining_tools.
"""

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

import biotite.structure.io as strucio
import biotite.structure as struc
import biotite.structure.io.pdbx as pdbx
import py3Dmol
from string import ascii_uppercase, ascii_lowercase
from openmm import unit
import openmm as mm
from openmm import app
from openmm.unit import *  # noqa: F403
from openmm.app import PDBFile
import tempfile
import os
import numpy as np


alphabet_list = list(ascii_uppercase + ascii_lowercase)
pymol_color_list = [
    "#33ff33",
    "#00ffff",
    "#ff33cc",
    "#ffff00",
    "#ff9999",
    "#e5e5e5",
    "#7f7fff",
    "#ff7f00",
    "#7fff7f",
    "#199999",
    "#ff007f",
    "#ffdd5e",
    "#8c3f99",
    "#b2b2b2",
    "#007fff",
    "#c4b200",
    "#8cb266",
    "#00bfbf",
    "#b27f7f",
    "#fcd1a5",
    "#ff7f7f",
    "#ffbfdd",
    "#7fffff",
    "#ffff7f",
    "#00ff7f",
    "#337fcc",
    "#d8337f",
    "#bfff3f",
    "#ff7fff",
    "#d8d8ff",
    "#3fffbf",
    "#b78c4c",
    "#339933",
    "#66b2b2",
    "#ba8c84",
    "#84bf00",
    "#b24c66",
    "#7f7f7f",
    "#3f3fa5",
    "#a5512b",
]

amino_acid_mapping = {
    "ALA": "A",
    "ARG": "R",
    "ASN": "N",
    "ASP": "D",
    "CYS": "C",
    "GLN": "Q",
    "GLU": "E",
    "GLY": "G",
    "HIS": "H",
    "ILE": "I",
    "LEU": "L",
    "LYS": "K",
    "MET": "M",
    "PHE": "F",
    "PRO": "P",
    "SER": "S",
    "THR": "T",
    "TRP": "W",
    "TYR": "Y",
    "VAL": "V",
    # Adding a generic mapping for unknowns
    "UNK": "X",  # Unknown
    "SEC": "U",  # Selenocysteine, sometimes considered the 21st amino acid
    # Additional mappings as needed
}


def load_struc(prot):
    """
    Load a protein structure.

    Args:
        prot biotite.structure.AtomArray or path to pdb (str).

    Returns:
        biotite.structure.AtomArray
    """
    if isinstance(prot, struc.AtomArray):
        prot = prot

    elif isinstance(prot, str):
        try:
            prot = strucio.load_structure(prot)
        except Exception as e:
            raise ValueError(f"prot 1 has an unexpected format. Error {e}")

    else:
        raise ValueError("pdb file has an unexpected format")

    return prot


def align(prot1, prot2):
    """
    Superimpose protein 2 on protein 1.

    Args:
        prot1 biotite.structure.AtomArray or path to pdb (str).
        prot2 biotite.structure.AtomArray or path to pdb (str).

    Returns:
        prot1, prot2_superimposed as biotite.structure.AtomArray's
    """
    prot1 = load_struc(prot1)
    prot2 = load_struc(prot2)

    prot1 = prot1[(prot1.chain_id == "A") | (prot1.chain_id == "B")]
    prot1 = prot1[~struc.filter_solvent(prot1)]
    prot2 = prot2[~struc.filter_solvent(prot2)]

    prot1_common = prot1[struc.filter_intersection(prot1, prot2)]
    prot2_common = prot2[struc.filter_intersection(prot2, prot1)]

    # Superimpose
    prot2_superimposed, transformation = struc.superimpose(
        prot1_common, prot2_common, (prot2_common.atom_name == "CA")
    )

    # We do not want the cropped structures
    prot2_superimposed = struc.superimpose_apply(prot2, transformation)

    # Write PDBx files as input for PyMOL
    cif_file = pdbx.PDBxFile()
    pdbx.set_structure(cif_file, prot1, data_block="prot1")
    cif_file.write("test1.cif")

    cif_file = pdbx.PDBxFile()
    pdbx.set_structure(cif_file, prot2_superimposed, data_block="prot2")
    cif_file.write("test2.cif")

    return prot1, prot2_superimposed


def compute_rmsd(prot1, prot2):
    """
    Compute rmsd between two proteins.

    Args:
        prot1 biotite.structure.AtomArray or path to pdb (str).
        prot2 biotite.structure.AtomArray or path to pdb (str).

    Returns:
        rmsd (float)
    """
    prot1 = load_struc(prot1)
    prot2 = load_struc(prot2)

    prot1_common = prot1[struc.filter_intersection(prot1, prot2)]
    prot2_common = prot2[struc.filter_intersection(prot2, prot1)]

    rmsd = struc.rmsd(prot1_common, prot2_common)
    print("{:.3f}".format(rmsd))
    return rmsd


def chain_parser(pdb_file):
    """
    Parse chains from pdb file.

    Args:
        pdb_file: path to pdb file (str) or AtomArray
    """

    prot = load_struc(pdb_file)

    chains = list(set(prot.chain_id))  # type: ignore
    return chains


def get_sequences(prot_f):
    """
    Get the sequence from a protein structure.

    Args:
        prot biotite.structure.AtomArray or path to pdb (str).

    Return:
        dictionary of chains and sequences
    """

    prot = load_struc(prot_f)

    chains = chain_parser(prot)

    sequences = {}
    for chain in chains:
        res_ids = list(set(prot[prot.chain_id == chain].res_id))
        residues = [
            amino_acid_mapping.get(prot[prot.res_id == r].res_name[0], "X")
            for r in res_ids
        ]
        sequences[chain] = "".join(residues)

    return sequences


def get_contacts(structure, chain=None, target="protein", dist=7.0):
    """
    Get contacts within a protein structure or between a protein and ligands, based on the specified chain.

    Args:
        structure (biotite.structure.AtomArray): The complete protein structure.
        chain (str, optional): Specific chain for which to compute the contacts.
            Default is None, which will use the first protein chain if not specified.
        target (str): Specify 'protein' for protein-protein contacts or 'ligand'
            for protein-ligand contacts. Default is 'protein'.
        dist (float): Specified distance threshold in Angstroms. Default is 7.

    Returns:
        list: Unique residue IDs in contact from the specified chain.
    """
    if chain is None:
        chain = structure.chain_id[0]  # Default to the first chain if none specified

    if target == "protein":
        # Select atoms from all other chains, excluding heteroatoms if needed
        target_atoms = structure[(structure.chain_id != chain) & (~structure.hetero)]
    elif target == "ligand":
        # Assuming ligands are identified as heteroatoms
        target_atoms = structure[structure.hetero]
    else:
        raise ValueError("Invalid target specified. Use 'protein' or 'ligand'.")

    # Select the atoms in the target chain, excluding heteroatoms if needed
    chain_atoms = structure[(structure.chain_id == chain) & (~structure.hetero)]

    # Initialize cell list with target atoms
    cell_list = struc.CellList(target_atoms, cell_size=dist)
    contact_indices = cell_list.get_atoms(chain_atoms.coord, radius=dist)

    # Determine contact residue IDs
    contact_residue_indices = np.where((contact_indices != -1).any(axis=1))[0]
    contact_res_ids = chain_atoms.res_id[contact_residue_indices]

    # Removing duplicate IDs by converting to a set, then back to sorted list
    unique_contact_res_ids = sorted(set(contact_res_ids))

    return unique_contact_res_ids


def compute_chi_angles(protein, res_ids):
    """
    Compute the chi angles for specified residues in a protein by chain.

    Args:
        protein: biotite.structure.AtomArray or path to pdb (str).
        res_ids (dict): Dictionary where keys are chain identifiers and values are lists of residue IDs
                        to compute chi angles for in each chain.
        chi_atom_names (dict): Dictionary where keys are residue names and values are lists of lists of
                               atom names involved in each chi angle.

    Returns:
        A dictionary where keys are tuples of (chain identifier, residue name, residue ID) and values
        are lists of chi angles.
    """

    # Dictionary containing atom names involved in chi angle calculations for each amino acid
    chi_atom_names = {
        "ALA": [],  # Alanine has no chi angles
        "ARG": [
            ["N", "CA", "CB", "CG"],
            ["CA", "CB", "CG", "CD"],
            ["CB", "CG", "CD", "NE"],
            ["CG", "CD", "NE", "CZ"],
        ],
        "ASN": [["N", "CA", "CB", "CG"], ["CA", "CB", "CG", "OD1"]],
        "ASP": [["N", "CA", "CB", "CG"], ["CA", "CB", "CG", "OD1"]],
        "CYS": [["N", "CA", "CB", "SG"]],
        "GLN": [
            ["N", "CA", "CB", "CG"],
            ["CA", "CB", "CG", "CD"],
            ["CB", "CG", "CD", "OE1"],
        ],
        "GLU": [
            ["N", "CA", "CB", "CG"],
            ["CA", "CB", "CG", "CD"],
            ["CB", "CG", "CD", "OE1"],
        ],
        "GLY": [],  # Glycine has no chi angles
        "HIS": [["N", "CA", "CB", "CG"], ["CA", "CB", "CG", "ND1"]],
        "ILE": [["N", "CA", "CB", "CG1"], ["CA", "CB", "CG1", "CD1"]],
        "LEU": [["N", "CA", "CB", "CG"], ["CA", "CB", "CG", "CD1"]],
        "LYS": [
            ["N", "CA", "CB", "CG"],
            ["CA", "CB", "CG", "CD"],
            ["CB", "CG", "CD", "CE"],
            ["CG", "CD", "CE", "NZ"],
        ],
        "MET": [
            ["N", "CA", "CB", "CG"],
            ["CA", "CB", "CG", "SD"],
            ["CB", "CG", "SD", "CE"],
        ],
        "PHE": [["N", "CA", "CB", "CG"], ["CA", "CB", "CG", "CD1"]],
        "PRO": [["N", "CA", "CB", "CG"], ["CA", "CB", "CG", "CD"]],
        "SER": [["N", "CA", "CB", "OG"]],
        "THR": [["N", "CA", "CB", "OG1"]],
        "TRP": [["N", "CA", "CB", "CG"], ["CA", "CB", "CG", "CD1"]],
        "TYR": [["N", "CA", "CB", "CG"], ["CA", "CB", "CG", "CD1"]],
        "VAL": [["N", "CA", "CB", "CG1"]],
    }

    if isinstance(protein, str):
        protein = load_struc(protein)  # Assuming load_struc loads a structure

    chi_angles = {}
    for chain, residues in res_ids.items():
        for res_id in residues:
            # Filter protein by chain and residue ID
            mask = (protein.chain_id == chain) & (protein.res_id == res_id)
            res_protein = protein[mask]
            res_name = res_protein.res_name[0] if len(res_protein) > 0 else None

            if res_name and res_name in chi_atom_names:
                chi_list = []
                for atom_names in chi_atom_names[res_name]:
                    if all(atom in res_protein.atom_name for atom in atom_names):
                        atoms = [
                            res_protein[res_protein.atom_name == atom][0]
                            for atom in atom_names
                        ]
                        chi = struc.dihedral(*atoms)
                        chi_list.append(chi)
                if chi_list:
                    chi_angles[(chain, res_name, res_id)] = chi_list

    return chi_angles


def delta_chi(chi_angles_1, chi_angles_2):
    """
    Compute the delta between two sets of chi angles.

    Args:
        chi_angles_1 (dict): Dictionary with keys as (residue name, residue ID)
                             and values as lists of chi angles from the first structure.
        chi_angles_2 (dict): Dictionary with keys as (residue name, residue ID)
                             and values as lists of chi angles from the second structure.

    Returns:
        float: Sum of absolute differences between corresponding chi angles.
    """
    total_difference = 0.0

    for key in chi_angles_1:
        if key in chi_angles_2:
            angles_1 = chi_angles_1[key]
            angles_2 = chi_angles_2[key]
            # Ensure both lists of angles have the same length
            if len(angles_1) == len(angles_2):
                for a1, a2 in zip(angles_1, angles_2):
                    total_difference += abs(a1 - a2)
            else:
                raise ValueError(f"Mismatch in number of chi angles for residue {key}.")
        else:
            raise ValueError(f"Residue {key} not found in both structures.")

    return total_difference


def show_pdb(
    pdb_path,
    color="confidence",
    vmin=50,
    vmax=90,
    chains=None,
    Ls=None,
    size=(800, 480),
    show_sidechains=False,
    show_mainchains=False,
    highlight=None,
    sticks=None,
):
    """
    This function displays the 3D structure of a protein from a given PDB file in a Jupyter notebook.
    The protein structure can be colored by chain, rainbow, pLDDT, or confidence value. The size of the
    display can be changed. The sidechains and mainchains can be displayed or hidden.
    Additional functionality includes highlighting specific residues within specific chains by passing
    a dictionary with chains as keys and lists of residue numbers as values, and displaying ligands
    as sticks and ions as spheres.
    Parameters:
        pdb_path (str): The filename of the PDB file that contains the protein structure.
        color (str, optional): The color scheme for the protein structure. Can be "chain", "rainbow", "pLDDT", "confidence". Defaults to "rainbow".
        vmin (float, optional): The minimum value of pLDDT or confidence value. Defaults to 50.
        vmax (float, optional): The maximum value of pLDDT or confidence value. Defaults to 90.
        chains (int, optional): The number of chains to be displayed. Defaults to None.
        Ls (list, optional): A list of the chains to be displayed. Defaults to None.
        size (tuple, optional): The size of the display window. Defaults to (800, 480).
        show_sidechains (bool, optional): Whether to display the sidechains. Defaults to False.
        show_mainchains (bool, optional): Whether to display the mainchains. Defaults to False.
        highlight (dict, optional): A dictionary with chains as keys and lists of residue numbers to be highlighted as values. Defaults to None.
        sticks (list, optional): A list of residues that should be displayed as sticks. Default to None.
    Returns:
        view: The 3Dmol view object that displays the protein structure.
    """

    with open(pdb_path) as ifile:
        system = "".join([x for x in ifile])

    view = py3Dmol.view(
        js="https://3dmol.org/build/3Dmol.js", width=size[0], height=size[1]
    )

    view.addModelsAsFrames(system)

    # Apply color styles based on function arguments
    if color == "pLDDT" or color == "confidence":
        view.setStyle(
            {},
            {
                "cartoon": {
                    "colorscheme": {
                        "prop": "b",
                        "gradient": "rwb",
                        "min": vmin,
                        "max": vmax,
                    }
                }
            },
        )
    elif color == "rainbow":
        view.setStyle({}, {"cartoon": {"color": "spectrum"}})
    else:
        # Set default style as white
        view.setStyle({}, {"cartoon": {"color": "white"}})

    # Highlight specific residues per chain
    if highlight and isinstance(highlight, dict):
        for chain, residues in highlight.items():
            for resi in residues:
                highlight_style = {
                    "stick" if resi in (sticks or []) else "cartoon": {
                        "colorscheme": "blueCarbon",
                        "radius": 0.3,
                    }
                }
                view.addStyle({"chain": chain, "resi": str(resi)}, highlight_style)

    if sticks:
        stick_style = {"stick": {"radius": 0.3}}
        for resi in sticks:
            view.addStyle({"resi": str(resi)}, stick_style)

    # Display ligands as sticks and ions as spheres
    view.addStyle(
        {"hetflag": True, "bonds": 0, "atom": "not O"},
        {"sphere": {"colorscheme": "ionic", "radius": 0.5}},
    )
    view.addStyle(
        {"hetflag": True}, {"stick": {"colorscheme": "organic", "radius": 0.3}}
    )  # Display organic ligands as sticks

    if show_sidechains:
        BB = ["C", "O", "N"]
        view.addStyle(
            {
                "and": [
                    {"resn": ["GLY", "PRO"], "invert": True},
                    {"atom": BB, "invert": True},
                ]
            },
            {"stick": {"colorscheme": "WhiteCarbon", "radius": 0.3}},
        )
        view.addStyle(
            {"and": [{"resn": "GLY"}, {"atom": "CA"}]},
            {"sphere": {"colorscheme": "WhiteCarbon", "radius": 0.3}},
        )
        view.addStyle(
            {"and": [{"resn": "PRO"}, {"atom": ["C", "O"], "invert": True}]},
            {"stick": {"colorscheme": "WhiteCarbon", "radius": 0.3}},
        )

    if show_mainchains:
        BB = ["C", "O", "N", "CA"]
        view.addStyle(
            {"atom": BB}, {"stick": {"colorscheme": "WhiteCarbon", "radius": 0.3}}
        )

    view.zoomTo()

    return view


def relax_pdb(file, dest="outputs/struc/relaxed"):
    """
    Processes and minimizes a protein structure file using molecular dynamics.

    The function takes a PDB file, corrects its structure by adding missing residues and atoms,
    and then performs an energy minimization. The minimization is carried out under specified
    conditions using a Langevin integrator and a given force field. If CUDA is available,
    the function attempts to use it for the computations; otherwise, it falls back to CPU.

    Parameters:
        file (str): The path to the PDB file to be processed.
        dest (str, optional): The directory where the relaxed PDB file will be saved.
                              Default is 'outputs/struc/relaxed'.

    Returns:
        str: The path to the relaxed PDB file, saved in the specified destination directory.

    Raises:
        EnvironmentError: If the CUDA platform is not available and the function falls back to CPU,
                           this is not an error per se but might be relevant for performance-sensitive applications.

    Example:
        relaxed_path = relax_pdb('input/protein.pdb')
        print(f'Relaxed structure saved to {relaxed_path}')

    Note:
        This function uses the Amber14 force field and assumes a pH of 7.0 for protonation states.
    """

    try:
        from pdbfixer import PDBFixer
    except Exception as e:
        raise ValueError(
            f"Relaxation of protein structures requires PDBFixer: Please install through conda:\nconda install conda-forge::pdbfixer. Error {e}"
        )

    name = file.split("/")[-1].split(".")[0]

    fixer = PDBFixer(filename=file)
    fixer.findMissingResidues()
    fixer.findMissingAtoms()
    fixer.addMissingAtoms()
    fixer.addMissingHydrogens(
        7.0
    )  # pH value to decide protonation state of HIS, ASP, GLU

    # Use a temporary file instead of a fixed file name
    with tempfile.NamedTemporaryFile(
        delete=False, suffix=".pdb", mode="w"
    ) as temp_file:
        PDBFile.writeFile(fixer.topology, fixer.positions, temp_file)
        temp_file_path = temp_file.name  # Store the temporary file name to use it later

    # Now read the temporary PDB file
    pdb = app.PDBFile(temp_file_path)

    # Prepare the force field
    forcefield = app.ForceField("amber14-all.xml", "amber14/tip3pfb.xml")

    # Create a system
    system = forcefield.createSystem(
        pdb.topology, nonbondedMethod=app.NoCutoff, constraints=app.HBonds
    )

    # Create an integrator
    integrator = mm.LangevinIntegrator(
        300 * unit.kelvin, 1.0 / unit.picoseconds, 2.0 * unit.femtoseconds
    )

    # Try to use CUDA, otherwise fallback to CPU
    try:
        platform = mm.Platform.getPlatformByName("CUDA")
        properties = {"CudaPrecision": "mixed"}
    except Exception:
        # Fallback to CPU if CUDA is not available
        platform = mm.Platform.getPlatformByName("CPU")
        properties = {}  # CPU does not need special properties

    # Create a simulation context
    simulation = app.Simulation(pdb.topology, system, integrator, platform, properties)

    # Set the initial positions
    simulation.context.setPositions(pdb.positions)

    # Minimize the energy
    simulation.minimizeEnergy()

    # Get the final positions after minimization
    positions = simulation.context.getState(getPositions=True).getPositions()

    # Save the minimized structure to a new PDB file
    relaxed_pdb = os.path.join(dest, f"{name}_relaxed.pdb")
    app.PDBFile.writeFile(pdb.topology, positions, open(relaxed_pdb, "w"))

    return relaxed_pdb



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

import biotite.sequence.graphics as graphics
import biotite.application.muscle as muscle
import matplotlib.pyplot as plt
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio.Align.Applications import ClustalwCommandline
from Bio import AlignIO
from Bio import SeqIO
from collections import Counter
from biotite.sequence import ProteinSequence


def align_proteins(
    names: list,
    seqs: list,
    plot_results: bool = False,
    plt_range: tuple = (0, 200),
    muscle_version: str = "5",
    save_fig: str = None,
    save_fasta: str = None,
    figsize: tuple = (10.0, 8.0),
):
    """
    performs multiple sequence alignement given a list of blast names and the corresponding sequences.

    Parameters:
        names (list): list of sequence names
        seqs (list): list of sequences for MSA
        plot_results (bool, optional): plot results.
        plt_range (tuple, optional): range of sequence which is plotted. Default first 200 amino acids.
        muscle_version (str, optional): which muscle version is installed on your machine. Default 5.
        save_fig (str): saves fig if path is provided. Default None - won't save the figure.
        save_fasta (str): saves fasta if path is provided. Default None - won't save fasta.
        figsize (tuple): dimensions of the figure.

    Returns:
        dict: MSA results of sequence names/ids and gapped sequences
    """

    # Convert sequences to ProteinSequence objects if they are strings
    seqs = [ProteinSequence(seq) if isinstance(seq, str) else seq for seq in seqs]

    if muscle_version == "5":
        app = muscle.Muscle5App(seqs)
    elif muscle_version == "3":
        app = muscle.MuscleApp(seqs)
    else:
        raise ValueError("Muscle version must be either 3 or 5")

    app.start()
    app.join()
    alignment = app.get_alignment()

    # Print the MSA with hit IDs
    gapped_seqs = alignment.get_gapped_sequences()

    MSA_results = {}
    for i in range(len(gapped_seqs)):
        MSA_results[names[i]] = gapped_seqs[i]

    # Reorder alignments to reflect sequence distance
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    order = app.get_alignment_order()
    graphics.plot_alignment_type_based(
        ax,
        alignment[plt_range[0] : plt_range[1], order.tolist()],
        labels=[names[i] for i in order],
        show_numbers=True,
        color_scheme="clustalx",
    )
    fig.tight_layout()

    if save_fig is not None:
        plt.savefig(save_fig)

    if plot_results:
        plt.show()
    else:
        plt.close()

    if save_fasta is not None:
        with open(save_fasta, "w") as f:
            for i, key in enumerate(MSA_results.keys()):
                s = MSA_results[key]
                if i < len(MSA_results) - 1:
                    f.writelines(f">{key}\n")
                    f.writelines(f"{s}\n")
                else:
                    f.writelines(f">{key}\n")
                    f.writelines(f"{s}")

    return MSA_results


def MSA_results_to_fasta(MSA_results: dict, fname: str):
    """
    Takes MSA results from the align proteins function and writes then into a fasta format.

    Parameters:
        MSA_results (dict): Dictionary of MSA results.
        fname (str): file name.

    Returns:
        None
    """
    with open(fname, "w") as f:
        for i, key in enumerate(MSA_results.keys()):
            s = MSA_results[key]
            if i < len(MSA_results) - 1:
                f.writelines(f">{key}\n")
                f.writelines(f"{s}\n")
            else:
                f.writelines(f">{key}\n")
                f.writelines(f"{s}")


def align_dna(dna_sequences: list, verbose: bool = False):
    """
    performs multiple sequence alignement given a list of blast names and the corresponding sequences.
    The function uses the clustalw2 app which uses the global needleman-wunsch algorithm.

    Parameters:
        seqs (list): list of DNA sequences for MSA
        verbose (bool): print std out and std error if True. Default False

    Returns:
        list: MSA list of sequences
    """
    # Create SeqRecord objects for each DNA sequence
    seq_records = [
        SeqRecord(Seq(seq), id=f"seq{i + 1}") for i, seq in enumerate(dna_sequences)
    ]

    # Write the DNA sequences to a temporary file in FASTA format
    temp_input_file = "temp_input.fasta"
    with open(temp_input_file, "w") as handle:
        SeqIO.write(seq_records, handle, "fasta")

    # Run ClustalW to perform the multiple sequence alignment
    temp_output_file = "temp_output.aln"
    clustalw_cline = ClustalwCommandline(
        "clustalw2", infile=temp_input_file, outfile=temp_output_file
    )
    stdout, stderr = clustalw_cline()

    if verbose:
        print("std out:")
        print("-------")
        print(stdout)
        print("std error:")
        print("----------")
        print(stderr)

    # Read in the aligned sequences from the output file
    aligned_seqs = []
    with open(temp_output_file) as handle:
        alignment = AlignIO.read(handle, "clustal")
        for record in alignment:
            aligned_seqs.append(str(record.seq))

    return aligned_seqs


def get_consensus_sequence(dna_sequences: list):
    """
    Calculates the consensus sequence of multiple sequence alignements.
    It uses the most common character of every sequence in a list. All
    sequences need to be the same length.

    Parameters:
        dna_sequences (list): list of DNA sequences.

    Returns:
        str: consensus sequence.
    """
    # Get the length of the sequences
    sequence_length = len(dna_sequences[0])

    # Iterate over each position in the sequences
    consensus_sequence = ""
    for i in range(sequence_length):
        # Create a list of the characters at this position
        char_list = [seq[i] for seq in dna_sequences]

        # Count the occurrences of each character
        char_counts = Counter(char_list)

        # Get the most common character
        most_common_char = char_counts.most_common(1)[0][0]

        # Add the most common character to the consensus sequence
        consensus_sequence += most_common_char

    return consensus_sequence



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

import numpy as np
import esm
import typing as T
import biotite.sequence as seq
import biotite.sequence.align as align
from biotite.structure.io.pdb import PDBFile
import biotite.structure as struc
from biotite.structure import sasa
import tempfile

from proteusAI.data_tools import pdb  # TODO: double check with Johny!


# _____Sequence Constraints_____
def length_constraint(seqs: list, max_len: int = 200):
    """
    Constraint for the length of a seuqences.

    Parameters:
        seqs (list): sequences to be scored
        max_len (int): maximum length that a sequence is allowed to be. Default = 300

    Returns:
        np.array: Energy values
    """
    energies = np.zeros(len(seqs))

    for i, sequence in enumerate(seqs):
        if len(sequence) > max_len:
            energies[i] = float(len(sequence) - max_len)
        else:
            energies[i] = 0.0

    return energies


def seq_identity(seqs, ref, matrix="BLOSUM62", local=False):
    """
    Calculates sequence identity of sequences against a reference sequence based on alignment.
    By default a global alignment is performed using the BLOSUM62 matrix.

    Parameters:
        seq1 (str): reference sequence
        seq2 (str): query sequence
        matrix (str): alignement matrix {BLOSUM62, BLOSUM50, BLOSUM30}. Default BLOSUM62
        local (bool): Local alignment if True, else global alignment.

    Returns:
        numpy.ndarray: identity scores of sequences
    """
    alph = seq.ProteinSequence.alphabet
    matrix = align.SubstitutionMatrix(alph, alph, matrix)

    seqs = [seq.ProteinSequence(s) for s in seqs]
    ref = seq.ProteinSequence(ref)

    scores = np.zeros(len(seqs))
    for i, s in enumerate(seqs):
        alignments = align.align_optimal(s, ref, matrix, local=local)
        score = align.get_sequence_identity(alignments[0])
        scores[i] = score

    return scores


# _____Structure Constraints_____
def string_to_tempfile(data):
    # create a temporary file
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        # write the string to the file
        temp_file.write(data.encode("utf-8"))
        # flush the file to make sure the data_tools is written
        temp_file.flush()
        # return the file object
        return temp_file


def create_batched_sequence_datasest(
    sequences: T.List[T.Tuple[str, str]], max_tokens_per_batch: int = 1024
) -> T.Generator[T.Tuple[T.List[str], T.List[str]], None, None]:
    """
    Taken from https://github.com/facebookresearch/esm/blob/main/scripts/esmfold_inference.py
    """
    batch_headers, batch_sequences, num_tokens = [], [], 0
    for header, sequence in sequences:
        if (len(sequence) + num_tokens > max_tokens_per_batch) and num_tokens > 0:
            yield batch_headers, batch_sequences
            batch_headers, batch_sequences, num_tokens = [], [], 0
        batch_headers.append(header)
        batch_sequences.append(sequence)
        num_tokens += len(sequence)

    yield batch_headers, batch_sequences


def structure_prediction(
    sequences: list,
    names: list,
    chunk_size: int = 124,
    max_tokens_per_batch: int = 1024,
    num_recycles: int = None,
):
    """
    Predict the structure of proteins.

    Parameters:
        sequences (list): all sequences for structure prediction
        names (list): names of the sequences
        chunck_size (int): Chunks axial attention computation to reduce memory usage from O(L^2) to O(L). Recommended values: 128, 64, 32.
        max_tokens_per_batch (int): Maximum number of tokens per gpu forward-pass. This will group shorter sequences together.
        num_recycles (int): Number of recycles to run. Defaults to number used in training 4.

    Returns:
        all_headers, all_sequences, all_pdbs, pTMs, mean_pLDDTs
    """
    model = esm.pretrained.esmfold_v1()
    model = model.eval().cuda()
    model.set_chunk_size(chunk_size)
    all_sequences = list(zip(names, sequences))

    batched_sequences = create_batched_sequence_datasest(
        all_sequences, max_tokens_per_batch
    )
    all_headers = []
    all_sequences = []
    all_pdbs = []
    pTMs = []
    mean_pLDDTs = []
    for headers, sequences in batched_sequences:
        output = model.infer(sequences, num_recycles=num_recycles)
        output = {key: value.cpu() for key, value in output.items()}
        pdbs = model.output_to_pdb(output)
        for header, sequence, pdb_string, mean_plddt, ptm in zip(
            headers, sequences, pdbs, output["mean_plddt"], output["ptm"]
        ):
            all_headers.append(header)
            all_sequences.append(sequence)
            all_pdbs.append(
                PDBFile.read(string_to_tempfile(pdb_string).name)
            )  # biotite pdb file name
            mean_pLDDTs.append(mean_plddt.item())
            pTMs.append(ptm.item())

    return all_headers, all_sequences, all_pdbs, pTMs, mean_pLDDTs


def globularity(pdbs):
    """
    globularity constraint

    Parameters:
        pdb (list): list of biotite pdb file

    Returns:
        np.array: variances of coordinates for each structure
    """
    variances = np.zeros(len(pdbs))

    for i, pdb_list in enumerate(pdbs):
        variance = pdb_list.get_coord().var()
        variances[i] = variance.item()

    return variances


def surface_exposed_hydrophobics(pdbs):
    """
    Calculate the surface exposed hydrophobics using the Shrake-Rupley (“rolling probe”) algorithm.

    Parameters:
        pdbs (list): list of biotite pdb file

    Returns:
        np.array: average sasa values for each structure
    """
    avrg_sasa_values = np.zeros(len(pdbs))
    for i, pdb_list in enumerate(pdbs):
        struc = pdb_list.get_structure()
        sasa_val = sasa(
            struc[0],
            probe_radius=1.4,
            atom_filter=None,
            ignore_ions=True,
            point_number=1000,
            point_distr="Fibonacci",
            vdw_radii="ProtOr",
        )
        sasa_mean = sasa_val.mean()
        avrg_sasa_values[i] = sasa_mean.item()

    return avrg_sasa_values


def backbone_coordination(samples: list, refs: list):
    """
    Superimpose structures and calculate the RMSD of the backbones.
    sample structures will be aligned against reference structures.

    Parameters:
    -----------
        samples (list): list of sample structures
        refs (list): list of reference structures

    Returns:
    --------
        np.array: RMSD values of alignments
    """
    if len(samples) != len(refs):
        raise "samples and refs must have the same length"

    rmsds = np.zeros(len(samples))

    for i in range(len(samples)):
        _, _, rmsd = pdb.struc_align(refs[i], samples[i])
        rmsds[i] = rmsd.item()

    return rmsds


def all_atom_coordination(samples, refs, sample_consts, ref_consts):
    """
    Calculate the RMSD of residues with all atom constraints.
    All atomic positions will be taken into consideration.

    Parameters:
    -----------
        samples (list): list of biotite pdb files (mutated)
        refs (list): list of biotite pdb files (reference)
        sample_consts (list): list of constraints on the sample
            (potentially shifts over time due to deletions and insertions)
        ref_consts (list): list of constraints of the reference

    Returns:
    --------
        np.array (len(samples),): calculated RMSD values for each sequence
    """

    rmsds = np.zeros(len(samples))

    for i in range(len(samples)):
        sample = samples[i]
        ref = refs[i]
        sample_const = sample_consts[i]
        ref_const = ref_consts[i]

        # get structures
        sample_struc = sample.get_structure()[0]
        ref_struc = ref.get_structure()[0]

        # get indices for alignment
        sample_indices = np.where(
            np.isin(sample_struc.res_id, [i + 1 for i in sample_const["all_atm"]])
        )
        ref_indices = np.where(
            np.isin(ref_struc.res_id, [i + 1 for i in ref_const["all_atm"]])
        )

        sample_struc_common = sample_struc[sample_indices[0]]
        ref_struc_common = ref_struc[ref_indices[0]]

        sample_superimposed, transformation = struc.superimpose(
            ref_struc_common, sample_struc_common
        )

        sample_superimposed = struc.superimpose_apply(sample_struc, transformation)

        sample_pdb = PDBFile()
        PDBFile.set_structure(sample_pdb, sample_superimposed)
        sample_coord = sample_pdb.get_coord()[0][sample_indices]

        ref_pdb = PDBFile()
        PDBFile.set_structure(ref_pdb, ref_struc)
        ref_coord = ref_pdb.get_coord()[0][ref_indices]

        rmsd = struc.rmsd(sample_coord, ref_coord)

        rmsds[i] = rmsd

    return rmsds



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

import requests
import os


def get_AF2_pdb(protein_id: str, out_path: str) -> bool:
    """
    This function takes in a UniProt ID and an output path and downloads the corresponding AlphaFold model
    from the EBI AlphaFold database in PDB format.

    Parameters:
    protein_id (str): The UniProt ID of the protein
    out_path (str): The path to save the PDB file to. The directory containing the output file will be created if it does not exist.

    Returns:
    bool: True if the PDB file was downloaded successfully, False otherwise.
    """
    os.makedirs(os.path.dirname(out_path), exist_ok=True)

    requestURL = f"https://alphafold.ebi.ac.uk/files/AF-{protein_id}-F1-model_v3.pdb"
    r = requests.get(requestURL)

    if r.status_code == 200:
        with open(out_path, "wb") as f:
            f.write(r.content)
            return True
    else:
        return False
